#!/usr/bin/env python 3.11.0
# -*-coding:utf-8 -*-
# @Author  : Shuang (Twist) Song
# @Contact   : SongshGeo@gmail.com
# GitHub   : https://github.com/SongshGeo
# Website: https://cv.songshgeo.com/

from __future__ import annotations

import os
import threading
from functools import cached_property
from numbers import Number
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    Self,
    Sequence,
    Tuple,
    Type,
    TypeAlias,
    Union,
)

import numpy as np
import pandas as pd
import rioxarray
import xarray as xr
from pyproj.crs.crs import CRS

from .tools.read_files import read_yaml

if TYPE_CHECKING:
    from abses.main import MainModel


NODATA = -9999.0
WGS84 = "EPSG:4326"

Coordinate: TypeAlias = Union[Iterable[Number], Dict[str, Iterable[Number]]]
Shape2: TypeAlias = Tuple[int, int]


class Geo:
    """Singleton AgentsContainer for each model."""

    _models: Dict[MainModel, Geo] = {}
    _lock = threading.RLock()

    def __new__(cls: type[Self], model: MainModel) -> Self:
        instance = cls._models.get(model, None)
        if instance is None:
            instance = super().__new__(cls)
            cls._initialize(instance, model)
            with cls._lock:
                cls._models[model] = instance
        return instance

    @classmethod
    def _initialize(cls, instance, model):
        instance._model = model
        instance._crs: CRS = None
        instance._nodata: float = None
        instance._dims: Tuple[str, str] = ("x", "y")
        instance._x: Optional[np.ndarray] = None
        instance._y: Optional[np.ndarray] = None
        instance._mask: Optional[np.ndarray] = None

    @property
    def shape(self) -> Tuple[int, int]:
        return self.height, self.width

    @property
    def mask(self) -> xr.DataArray:
        if self._mask is None:
            return self.zeros()
        else:
            return self.wrap_data(self._mask)

    @property
    def accessible(self) -> xr.DataArray:
        return ~self.mask

    @property
    def crs(self) -> CRS:
        return self._crs

    @crs.setter
    def crs(self, crs: Union[str, CRS, None]):
        if crs is not None:
            crs = CRS.from_user_input(crs)
        self._crs = crs

    @property
    def nodata(self) -> float:
        return self._nodata

    @nodata.setter
    def nodata(self, value: float):
        self._nodata = value

    @property
    def dims(self) -> Tuple[str, str]:
        return self._dims

    @dims.setter
    def dims(self, dims: Tuple[str, str]):
        self._dims = dims

    @property
    def geographic_crs_name(self) -> str:
        if self.crs is None:
            return "Not exists"
        else:
            return self.crs.name

    @property
    def georef(self) -> Dict[str, Any]:
        georef = {
            "crs": f"{self.geographic_crs_name}",
            "dims": f"{self.dims}",
            "nodata": f"{self.nodata}",
        }
        return georef

    @property
    def width(self) -> int:
        return len(self._x)

    @property
    def height(self) -> int:
        return len(self._y)

    @property
    def x(self) -> np.ndarray:
        return self._x

    @property
    def y(self) -> np.ndarray:
        return self._y

    @property
    def coords(self) -> Dict[str, np.ndarray]:
        coords = {
            self.dims[1]: self.y,
            self.dims[0]: self.x,
        }
        return coords

    def _setup_mask(self, value):
        if not hasattr(value, "shape"):
            raise AttributeError("Mask must has 'shape' attribute")
        elif value.shape != self.shape:
            raise ValueError(
                f"Input shape {value.shape} does not match {self.shape}."
            )
        self._mask = value

    def show_georef(self) -> pd.Series:
        return pd.Series(self.georef)

    def retrieve_georef(self, **kwargs) -> None:
        self.crs = kwargs.pop("crs", WGS84)
        self._nodata = kwargs.pop("nodata", NODATA)

    def setup_from_coords(
        self,
        coord_x: Coordinate,
        coord_y: Coordinate,
        **kwargs: Dict[str, Any],
    ) -> None:
        """
        Setup geographic references from coordinates.

        Args:
            coords (Tuple[int, int]): a dictionary of coordinates. Must be 2-dimensions coords. The following notations are accepted:
                    - mapping {dimension name: array-like}
                    - sequence of tuples that are valid arguments for
                    ``xarray.Variable()``
                    - (dims, data)
                    - (dims, data, attrs)
                    - (dims, data, attrs, encoding)
                If a list, it should be a list of tuples where the first element is the dimension name and the second element is the corresponding coordinate array_like object. see documentation of `Xarray` https://docs.xarray.dev/en/stable/user-guide/data-structures.html.

        Returns:
            Tuple[int, int]: shape of the array.
        """

        def parse_coord(coord, default):
            if isinstance(coord, dict):
                name = coord.get("name", default)
                data = coord["data"]
            else:
                name = default
                data = np.array(coord)
                if len(data.shape) != 1:
                    raise TypeError
            return name, data

        x, self._x = parse_coord(coord_x, "x")
        y, self._y = parse_coord(coord_y, "y")
        self._dims = (x, y)
        self.retrieve_georef(**kwargs)

    def setup_from_shape(
        self,
        shape: Tuple[int, int],
        resolution: Union[int, float] = 1,
        **kwargs,
    ) -> None:
        height, width = shape
        self._x = np.arange(0, width * resolution, resolution)
        self._y = np.arange(0, height * resolution, resolution)
        self.retrieve_georef(**kwargs)

    @staticmethod
    def detect_dims(dims: Tuple[str]) -> Tuple[str]:
        if "x" in dims and "y" in dims:
            return ("x", "y")
        elif "lon" in dims and "lat" in dims:
            return ("lon", "lat")
        elif "longitude" in dims and "latitude" in dims:
            return ("longitude", "latitude")
        else:
            raise ValueError("Unknown dimensions %s." % dims)

    def setup_from_dict(self, settings: dict) -> None:
        shape = settings.pop("shape", None)
        resolution = settings.pop("resolution", 1)
        if shape is None:
            try:
                width = settings.pop("width")
                height = settings.pop("height")
            except KeyError:
                raise KeyError(
                    f"'shape' or ('width' and 'height') not specified in {settings.keys()}!"
                )
            shape = (height, width)
        self.setup_from_shape(shape=shape, resolution=resolution, **settings)

    def setup_from_file(self, filename: str) -> None:
        path = Path(filename)
        if not path.is_file():
            raise ValueError("Could not find file %s" % filename)
        if path.suffix in [".tiff", ".tif"]:
            xda = rioxarray.open_rasterio(filename)
            dims = self.detect_dims(xda.dims)
            x_coord = xda.coords[dims[0]].to_dict()
            y_coord = xda.coords[dims[1]].to_dict()
            self.setup_from_coords(x_coord, y_coord, **xda.attrs)
            nodata = xda.isnull().data.reshape(self.shape)
            self._setup_mask(nodata)
        elif path.suffix in [".yaml"]:
            settings = read_yaml(path)
            self.setup_from_dict(settings)
        else:
            raise ValueError(f"Unknown referring file type {path.suffix}.")

    def auto_setup(self, settings: Union[str, dict]) -> None:
        if isinstance(settings, str):
            self.setup_from_file(settings)
        elif isinstance(settings, dict):
            self.setup_from_dict(settings)
        else:
            raise TypeError(f"Not support '{type(settings)}'.")

    def zeros(self, dtype: Type = bool) -> xr.DataArray:
        data = np.zeros(self.shape, dtype=dtype)
        return self.wrap_data(data, masked=False)

    def ones(self, dtype: Type = bool) -> xr.DataArray:
        data = np.ones(self.shape, dtype=dtype)
        return self.wrap_data(data, masked=False)

    def wrap_data(
        self, data: Optional[np.ndarray] = None, masked: bool = False
    ) -> xr.DataArray:
        if data is None:
            return self.accessible
        xda = xr.DataArray(data, coords=self.coords)
        xda.rio.write_crs(self.crs, inplace=True)
        if masked is True:
            return xda.where(self.accessible)
        else:
            return xda

    def clip_match(self, xda: xr.DataArray) -> xr.DataArray:
        xda = xda.rio.write_crs(self.crs)
        aligned = xda.rio.reproject_match(self.accessible)
        return aligned



#!/usr/bin/env python 3.11.0
# -*-coding:utf-8 -*-
# @Author  : Shuang Song
# @Contact   : SongshGeo@gmail.com
# GitHub   : https://github.com/SongshGeo
# Website: https://cv.songshgeo.com/

from copy import deepcopy
from functools import cached_property
from pathlib import Path
from typing import List, Optional, Tuple, Union

import numpy as np
import xarray
from agentpy import AttrDict
from agentpy.grid import AgentSet

from abses.actor import Actor
from abses.bases import Creation, Creator
from abses.boundary import Boundaries
from abses.engine import GeoEngine
from abses.geo import Geo

from .algorithms.spatial import points_to_polygons, polygon_to_mask
from .container import AgentsContainer
from .modules import CompositeModule, Module
from .patch import Patch, get_buffer
from .sequences import ActorsList, Selection
from .tools.func import norm_choice

DEFAULT_WORLD = {
    "width": 9,
    "height": 9,
    "resolution": 10,
    # 'units': 'm',
}


class PositionSet(AgentSet):
    def __init__(self, model, index, accessible, *args, **kwargs):
        super().__init__(model, *args, **kwargs)
        self._index: Tuple[int, int] = index
        self._accessible: bool = accessible

    @property
    def index(self):
        return self._index

    @property
    def accessible(self):
        return self._accessible

    def add(self, actor: Actor):
        if self._accessible is False:
            raise KeyError(f"{self.index} is not accessible.")
        else:
            super().add(actor)


class PatchModule(Module, Creator, Creation):
    _valid_type = (bool, int, float, str, "float32")
    _valid_dtype = tuple([np.dtype(t) for t in _valid_type])

    def __init__(self, model, name=None, **kwargs):
        Module.__init__(self, model, name=name)
        Creator.__init__(self)
        self._geo = Geo(model)
        self._mask = None
        # Other attrs
        self._attrs = kwargs.copy()
        self._patches = AttrDict()

    @property
    def geo(self):
        return self._geo

    @property
    def mask(self) -> xarray.DataArray:
        if self._mask is None:
            self._mask = self.geo.zeros(bool)
        mask = self._mask | self.geo.mask
        return mask.rio.write_crs(self.geo.crs)

    @property
    def accessible(self):
        return ~self.mask

    @property
    def attrs(self):
        return self._attrs

    def _check_dtype(self, values) -> None:
        dtype = values.dtype
        if dtype not in self._valid_dtype:
            raise TypeError(f"Invalid value type {dtype}.")

    def _check_type(self, value) -> type:
        val_type = type(value)
        if val_type not in self._valid_type:
            raise ValueError(f"Invalid type {val_type}")

    def _check_shape(self, values):
        if values.shape != self.geo.shape:
            raise ValueError(
                f"Invalid shape {values.shape}, mismatch with shape {self.shape}."
            )

    def create_patch(
        self,
        values: Union[np.ndarray, str, bool, float, int],
        name: str,
        xarray: bool = True,
        add: bool = False,
    ) -> Patch:
        if not hasattr(values, "shape"):
            # only int|float|str|bool are supported
            self._check_type(values)
            values = np.full(self.geo.shape, values)
        else:
            # nd-array like data
            self._check_dtype(values)
            self._check_shape(values)
        patch = Patch(values, name=name, father=self, xarray=xarray)
        self.add_creation(patch)
        if add is True:
            self._patches[name] = patch
        return patch

    @property
    def patches(self):
        return self._patches

    # @patches.setter
    # def patches(self, patch_name: str) -> None:
    #     self.creator.transfer_var(self, patch_name)
    #     self._patches.append(patch_name)

    # @property
    # def num_attrs(self):
    #     return self._num_attrs

    # @property
    # def bool_attrs(self):
    #     return self._bool_attrs

    def init_variables(self):
        # Hydraulic attributions.
        for attr in self.num_attrs:
            value = self.params.get(attr, 0.0)
            self.create_patch(value, attr, add=True)

        # Type mask with bool dtype.
        for attr in self.bool_attrs:
            value = self.params.get(attr, False)
            self.create_patch(False, attr, add=True)

    def read_patch(self, path, name) -> None:
        path = GeoEngine(path, self.model)
        # dataset = path.get_dataset()
        array = path.read2array()
        patch = self.create_patch(array, name=name)
        self._patches[name] = patch
        return patch
        # self.register_a_var(name=patch.name, value=patch.value)
        # self.patches = patch.name
        # setattr(self, patch.name, patch)

    # def get_patch(self, attr):
    #     return getattr(self, attr)

    # def update_patch(
    #     self,
    #     patch_name: str,
    #     value: "str|int|float|bool|np.ndarray",
    #     mask: np.ndarray = None,
    # ):
    #     if patch_name in self.patches:
    #         self.logger.warning(
    #             f"{patch_name} was created by this module, use 'patch.update()' method instead."
    #         )
    #     else:
    #         self.mediator.transfer_update(
    #             self, patch_name, value=value, mask=mask
    #         )
    #     pass


class BaseNature(CompositeModule, PatchModule):
    def __init__(self, model, name="nature", **kwargs):
        PatchModule.__init__(self, model=model, **kwargs)
        CompositeModule.__init__(self, model, name=name)
        self._boundary: Boundaries = None
        self._grid: np.ndarray = None

    # @property
    # def patches(self):
    #     return tuple(self._patches.keys())

    def __getitem__(self, key: Union[Tuple[int, int], slice]) -> ActorsList:
        items = self.grid[key]
        if isinstance(items, AgentSet):
            return ActorsList(self.model, items)
        else:
            return self._aggregate_agents(items)

    @property
    def boundary(self) -> Boundaries:
        return self._boundary

    @property
    def grid(self) -> np.ndarray:
        return self._grid

    def _aggregate_agents(self, items: np.ndarray) -> ActorsList:
        """Aggregating searched `PositionSet`s into an `ActorsList`."""
        agents = ActorsList(self.model)
        for item in items.flatten():
            agents.extend(item)
        return agents

    def _setup_grid(self, shape: Tuple[int, int]):
        """A numpy 2-d Grid where agents are saved in."""
        array = np.empty(shape=shape, dtype=object)
        access = self.accessible.to_numpy()
        it = np.nditer(array, flags=["refs_ok", "multi_index"])
        for _ in it:
            index = it.multi_index
            array[index] = PositionSet(
                model=self.model, accessible=access[index], index=index
            )
        self._grid = array

    def _after_parsing(self):
        """After parsing parameters, setup grid and geographic settings."""
        settings = deepcopy(self.params.get("world", DEFAULT_WORLD))
        self.geo.auto_setup(settings=settings)
        boundary_settings = self.params.get("boundary", {})
        self._boundary = Boundaries(shape=self.geo.shape, **boundary_settings)
        self._setup_grid(shape=self.geo.shape)

    def get_patch(self, attr: str, **kwargs) -> Patch:
        # TODO: finish this method
        if attr in self.patches:
            patch = self._patches[attr]
            return patch
        for module in self.modules:
            if attr in module.patches:
                return module._patches[attr]
        else:
            raise ValueError(f"Unknown patch {attr}.")

    def actor_to(self, actor: Actor, position: Tuple[int, int]):
        # TODO: refactor this function
        if actor.on_earth is True:
            self.grid[actor.pos].remove(actor)
        self.grid[position].add(actor)

    # def transfer_var(self, sender: object, var: str) -> None:
    #     if var not in self.patches:
    #         self._patches[var] = sender
    #     else:
    #         module = self._patches[var]
    #         if sender is module:
    #             self.logger.warning(
    #                 f"Transfer exists {var} of {module}, please use 'update' function to do so."
    #             )
    #         self.logger.error(
    #             f"{sender} wants to transfer an exists var '{var}', which was created by {module}!"
    #         )

    # def transfer_update(
    #     self, sender: object, patch_name: str, **kwargs
    # ) -> None:
    #     # update a patch
    #     if patch_name in self.patches:
    #         owner = self._patches[patch_name]
    #         getattr(owner, patch_name).update(**kwargs)
    #         self.logger.debug(
    #             f"{sender} requires {patch_name} of {owner} to update."
    #         )

    def random_positions(
        self,
        k: int,
        where: np.ndarray = None,
        probabilities: np.ndarray = None,
        only_empty: bool = False,
        replace: bool = False,
    ) -> List[Tuple[int, int]]:
        """
        Choose 'k' patches in the world randomly.

        Args:
            k (int): number of patches to choose.
            mask (np.ndarray, optional): bool mask, only True patches can be choose. If None, all patches are accessible. Defaults to None.
            replace (bool, optional): If a patch can be chosen more than once. Defaults to False.

        Returns:
            List[Tuple[int, int]]: iterable coordinates of chosen patches.
        """
        if where is None:
            where = self.accessible
        else:
            where = self.accessible & where
        if only_empty is True:
            where = where & ~self.has_agent().astype(bool)
        where = self.create_patch(where, "where")
        potential_pos = [(x, y) for x, y in where.arr.where()]
        if probabilities is not None:
            probabilities = probabilities[where]
        pos_index = norm_choice(
            np.arange(len(potential_pos)),
            size=k,
            p=probabilities,
            replace=replace,
        )
        positions = [potential_pos[i] for i in pos_index]
        return positions

    def add_agents(self, agents: ActorsList, positions=None, **kwargs):
        if positions is None:
            positions = self.random_positions(len(agents), **kwargs)
        for actor, pos in zip(agents, positions):
            actor.settle_down(pos)
        # msg = f"Randomly placed {len(agents)} '{agents.breed()}' in nature."
        # self.mediator.transfer_event(self, msg)

    def land_allotment(
        self, agents: ActorsList, where: np.ndarray = None
    ) -> AgentsContainer:
        """
        > For each cell in the grid, find the nearest agent and assign that agent's id to the cell

        :param mask: a boolean array that indicates which cells should be assigned an owner
        :return: The pattern of the agents.
        """
        where = self.geo.wrap_data(where, masked=True)
        points = agents.array("pos")
        polygons = points_to_polygons(points)
        for i, agent in enumerate(agents):
            owned_land = polygon_to_mask(polygons[i], shape=self.shape)
            owned_land = owned_land & where
            agent.attach_places(name="owned", place=owned_land)

    def lookup_agents(
        self, where: np.ndarray, selection: Optional[Selection] = None
    ) -> ActorsList:
        """
        Search agents.

        Args:
            where (np.ndarray): bool mask, when a cell is False, ignore agents here.
            selection (Optional[Selection], optional): filter results after search. see `ActorsList.select()` for more information. Defaults to None, meaning that no further selection, just returns the actual agents distribution.

        Returns:
            ActorsList: all qualified agents.
        """
        where = self.geo.wrap_data(where, masked=True)
        where_bool = where.to_numpy().astype(bool)
        agents = self._aggregate_agents(self.grid[where_bool])
        if selection is None:
            return agents
        else:
            return agents.select(selection)

    def neighbors(
        self,
        pos: Tuple[int, int],
        distance: int = 0,
        approach: int = 4,
        selection: Optional[Selection] = None,
    ) -> ActorsList:
        """
        Find all agents nearby with a given position and distance.

        Args:
            pos (Tuple[int, int]): a specific position in the world.
            distance (int, optional): distances to search. Defaults to 0.
            neighbors (int, optional): neighbor rule. Defaults to 4.
            breed (str, optional): filter results after search. see `ActorsList.select()` for more information. Defaults to None, meaning that no further selection, just returns the actual agents distribution.

        Returns:
            ActorsList: all qualified agents.
        """
        position = self.geo.zeros()
        position[pos] = True
        buffer = get_buffer(
            array=position, buffer=distance, neighbors=approach
        )
        return self.lookup_agents(buffer, selection=selection)

    def has_agent(self, selection: Optional[Selection] = None) -> Patch:
        """
        How many qualified agents are available in each cell.

        Args:
            selection (Optional[Selection], optional): selection, see `ActorsList.select()` for more information. Defaults to None, meaning that no further selection, just returns the actual agents distribution.

        Returns:
            Patch[int]: number of qualified agents in each cell.
        """

        def counts_agent(agents, selection):
            agents = ActorsList(self.model, agents)
            if selection is not None:
                return len(agents.select(selection))
            else:
                return len(agents)

        has_agents = np.vectorize(counts_agent)(self.grid, selection)
        patch = self.create_patch(has_agents, "has_agent", True)
        return patch



#!/usr/bin/env python 3.11.0
# -*-coding:utf-8 -*-
# @Author  : Shuang Song
# @Contact   : SongshGeo@gmail.com
# GitHub   : https://github.com/SongshGeo
# Website: https://cv.songshgeo.com/

from __future__ import annotations

import logging
from collections import namedtuple
from typing import (
    Dict,
    Iterable,
    NamedTuple,
    Optional,
    Tuple,
    TypeAlias,
    Union,
)

from agentpy import AttrDict, Model

from abses import __version__
from abses.tools.read_files import read_yaml

from .actor import Actor
from .bases import Mediator, Notice
from .components import STATES, MainComponent
from .container import AgentsContainer
from .human import BaseHuman
from .log import Log
from .nature import BaseNature
from .project import Folder
from .sequences import ActorsList
from .time import TimeDriver
from .tools.func import iter_func, make_list
from .variable import VariablesRegistry

logger = logging.getLogger(__name__)
LENGTH = 40  # session fill


class MainModel(Folder, MainComponent, Model, Notice):
    def __init__(
        self,
        name: Optional[str] = None,
        base: Optional[str] = None,
        parameters: Optional[Dict[str, any]] = None,
        human_class: Optional[BaseHuman] = None,
        nature_class: Optional[BaseNature] = None,
        settings_file: Optional[str] = None,
        _run_id: Optional[Tuple[int, int]] = None,
        **kwargs,
    ) -> None:
        Model.__init__(self, _run_id=_run_id)
        MainComponent.__init__(self, name=name)
        Folder.__init__(self, base=base)
        Notice.__init__(self)

        if nature_class is None:
            nature_class = BaseNature
        if human_class is None:
            human_class = BaseHuman
        if parameters is None:
            parameters = {}

        self.__version__: str = __version__
        self._human: BaseHuman = human_class(self)
        self._nature: BaseNature = nature_class(self)
        self._agents = AgentsContainer(model=self)
        # parameters
        # priority: init parameters > input parameters > settings_file
        self._init_params: AttrDict = AttrDict()
        self._init_params.update(parameters)
        self._init_params.update(kwargs)

        if settings_file is None:
            settings_file = parameters.pop("settings_file", None)
        self._settings_file: Optional[str] = settings_file
        # setup mediator
        self.mediator = MainMediator(
            model=self, human=self.human, nature=self.nature
        )
        self._registry = VariablesRegistry(model=self)
        self._time = TimeDriver(model=self)

    # ----------------------------------------------------------------
    def __repr__(self):
        name = self.__class__.__name__
        version = self.__version__
        return f"<{name}-{version}({self.name}): {self.state}>"

    # ----------------------------------------------------------------
    @property
    def agents(self) -> AgentsContainer:
        return self._agents

    @property
    def all_agents(self) -> ActorsList:
        return self.agents.to_list()

    @property
    def human(self) -> BaseHuman:
        return self._human

    @property
    def nature(self) -> BaseNature:
        return self._nature

    @property
    def registry(self) -> VariablesRegistry:
        return self._registry

    @property
    def time(self) -> TimeDriver:
        return self._time

    @property
    def _settings_from_file(self) -> Dict[str, any]:
        # settings yaml as basic
        settings = AttrDict()
        if self._settings_file is not None:
            settings = read_yaml(self._settings_file, nesting=True)
        return settings

    @property
    def settings(self) -> Dict[str, any]:
        settings = AttrDict()
        # read basic settings file
        settings.update(self._settings_from_file)
        # input settings
        settings.update(self._init_params)
        return settings

    @iter_func("observers")
    def _when_time_go(self):
        pass

    def time_go(self, steps: int = 1) -> TimeDriver:
        for _ in range(steps):
            self.time.update()
            self._when_time_go()
            self.t += 1
            self.step()
            self.update()
        return self.time

    def sim_step(self):
        """Proceeds the simulation by one step, incrementing `Model.t` by 1
        and then calling :func:`Model.step` and :func:`Model.update`."""
        self.time_go()
        if self.t >= self._steps:
            self.running = False


Sender: TypeAlias = Union[MainComponent, Actor]
TypingUsers: TypeAlias = NamedTuple(
    "Users",
    [("human", BaseHuman), ("nature", BaseNature), ("model", MainModel)],
)
TypingResults: TypeAlias = NamedTuple(
    "Users", [("human", any), ("nature", any), ("model", any)]
)
USERS: Tuple[str] = ("model", "human", "nature")
Users: TypingUsers = namedtuple("Users", USERS)


class MainMediator(Mediator, Log):
    def __init__(self, model: MainModel, human: BaseHuman, nature: BaseNature):
        self.model: MainModel = model
        self.human: BaseHuman = human
        self.nature: BaseNature = nature
        model.mediator: MainMediator = self
        human.mediator: MainMediator = self
        nature.mediator: MainMediator = self
        self.users: TypingUsers = Users(
            model=model, human=human, nature=nature
        )
        self.sender: Optional[Sender] = None
        name = ", ".join([user.name for user in self.users])
        Log.__init__(self, name=name)
        self._change_state(0)  # state -> new: initialize all

    def __repr__(self) -> str:
        return f"<Mediator of: {self.name}>"

    def _change_state(self, state_code: int) -> bool:
        f"""
        Change all users' state to the assigned state.

        Args:
            state_code (int): {STATES}.

        Raises:
            ValueError: Invalid state code.
        """
        if state_code not in STATES:
            raise ValueError(
                f"Invalid state code {state_code}, valid: {STATES}"
            )
        for user in self.users:
            user.state = state_code
        return self._states_are(STATES[state_code])

    def _check_sender(self, sender: Sender) -> Dict[Sender, bool]:
        """
        Check the type of sender, save it as a string attribute.
        An available pattern includes:
            1. 'agent': any instance of 'Actor'.
            2. 'model': the bound instance of 'MainModel'.
            3. 'human': the bound instance of 'BaseHuman'.
            4. 'nature': the bound instance of 'BaseNature'.

        Args:
            sender (Sender): the sending request object.
        """
        if sender is self.model:
            self.sender = "model"
        elif sender is self.human:
            self.sender = "human"
        elif sender is self.nature:
            self.sender = "nature"
        elif isinstance(sender, Actor):
            self.sender = "agent"
        else:
            raise TypeError(f"Type of sender '{type(sender)}' is invalid.")

    def sender_matches(self, *args: str) -> bool:
        """
        Check if the sender now matches ANY given pattern.
        An available pattern includes:
            1. 'agent': any instance of 'Actor'.
            2. 'model': the bound instance of 'MainModel'.
            3. 'human': the bound instance of 'BaseHuman'.
            4. 'nature': the bound instance of 'BaseNature'.

        Returns:
            bool: if the 'self.sender' matches any input pattern, returns True.
        """
        is_matched = [self.sender == arg for arg in args]
        return any(is_matched)

    def session(self, msg: str, sep: str = ".", new_line: int = 0) -> str:
        """
        Wrap a session message.

        Args:
            msg (str): user's msg to wrap.
            sep (str, optional): separator. Defaults to ".".
            new_line (int, optional): how many new lines '\n' after wrapped message. Defaults to 0.

        Returns:
            str: a wrapped message.
        """
        wrapped_msg = " [%s] ".center(LENGTH, sep) % msg + "\n" * new_line
        self.logger.info(wrapped_msg)
        return wrapped_msg

    def _states_are(self, state: str, how: str = "all") -> bool:
        f"""
        Check users' states are same as the given state.

        Args:
            state (str): checking states {STATES}:
            how (str, optional):
                if 'all', returns True only if all users' state are checked.
                if 'any', any user's state checked, returns True.
                Defaults to 'all'.

        Raises:
            ValueError: Invalid input args.

        Returns:
            bool: if the check is successful.
        """
        states = [user.state == state for user in self.users]
        if how == "all":
            return all(states)
        elif how == "any":
            return any(states)
        else:
            raise ValueError(f"{how} in invalid, choose 'any' or 'all'.")

    def trigger_functions(
        self, users: Union[str, Iterable[str]], func_name: str, *args, **kwargs
    ) -> TypingResults:
        f"""
        Trigger functions for all users

        Args:
            users (Union[str, Iterable[str]]): which users to trigger, choose one or more from {USERS}.
            func_name (str): name of the function to trigger.
            *args, **kwargs: additional arguments to pass to the func.

        Returns:
            results (NamedTuple[({USERS}), Optional[any]]): a named tuple where each user in {USERS} storing their results.
        """
        results = {}
        for user in USERS:
            if user not in make_list(users):
                results[user] = None
                continue
            obj = self.users.__getattribute__(user)
            func = obj.__getattribute__(func_name)
            results[user] = func(*args, **kwargs)
        return Users(**results)

    def logging(self, message, condition: bool = True, level="warning"):
        logging = getattr(self.logger, level)
        if condition:
            logging(message)

    def transfer_parsing(self, sender: Sender, params: Dict[str, any]) -> bool:
        f"""
        Transfer parameters parsing.

        Args:
            sender (Sender): any component from {USERS}.
            params (AttrDict): dictionary of parameters to parse.

        Returns:
            bool: parsing parameters finished if all model, human, nature components finished, returns True, else False.
        """
        self._check_sender(sender)
        all_finished = False
        # MainModel parsing parameters finished, trigger others.
        if self.sender_matches("model"):
            self.model.p.update(
                self.model.params
            )  # todo remove this if self.p is removed
            self.human.parsing_params(params)
            self.nature.parsing_params(params)
            return all_finished
        # Human / Nature parsing parameters finished, report.
        elif self.sender_matches("human", "nature"):
            sender.state = 1
            self.logger.info(f"{sender.name} parsed parameters.")
            # finished parsing parameters, change state to initialized.
            if self._states_are("init", how="all"):
                self.logger.info("All parameters are initialized.")
                all_finished = True
            return all_finished

    def _new(self):
        if self.sender_matches("model"):
            # TODO run id
            run_id = self.model._run_id
            if run_id is not None:
                run_id = run_id[0]
            self.session(f"Model {self.model.name} ID-{run_id}", sep="*")
        elif self.sender_matches("human", "nature"):
            pass
        if self._states_are("new"):
            # Automatically parsing parameters
            self.model.state = 1

    def _init(self):
        if self.sender_matches("model"):
            self.session("Parsing parameters")
            self.model.parsing_params(self.model.settings)
            self.model.initialize()
            self.nature.initialize()
            self.human.initialize()
            self.session("Initialized")
        elif self.sender_matches("human", "nature"):
            pass

    def _ready(self):
        if self.sender_matches("model"):
            self.session("Ready for simulation")

    def _complete(self):
        if self.sender_matches("model"):
            self.session(f"Completed in {self.model.t} steps", new_line=0)
            self.session("Finished", sep="*", new_line=1)
            self.human.state = 3
            self.nature.state = 3
        if self.sender_matches("human"):
            self.human.report_vars()
        if self.sender_matches("nature"):
            self.nature.report_vars()

    def transfer_event(
        self, sender: object, event: str, *args, **kwargs
    ) -> None:
        self._check_sender(sender)
        event_func = self.__getattribute__(f"_{event}")
        return event_func(*args, **kwargs)

    def transfer_request(self, sender: object, attr: str, **kwargs) -> object:
        self._check_sender(sender)
        if self.sender_matches("agent"):
            response = self.nature.get_patch(attr, **kwargs)
        else:
            response = None
        return response



#!/usr/bin/env python 3.11.0
# -*-coding:utf-8 -*-
# @Author  : Shuang (Twist) Song
# @Contact   : SongshGeo@gmail.com
# GitHub   : https://github.com/SongshGeo
# Website: https://cv.songshgeo.com/

import pathlib
from typing import Callable, Union

import netCDF4
import numpy as np
import rioxarray
import xarray

from abses.geo import Geo
from abses.time import TimeDriver


class GeoEngine:
    def __new__(cls, path, model):
        instance = super().__new__(cls)
        instance.path = pathlib.Path(path)
        instance.model = model
        instance.geo = model.nature.geo
        instance.time = model.time
        return instance

    # def __init__(self, path, geo, time):
    #     super().__init__(self, path)
    #     self.geo = geo
    #     self.time = time

    def read2array(
        self,
        resample_method: Union[str, Callable] = "mean",
        selecting_method: str = "nearest",
    ) -> np.ndarray:
        if self.path.suffix in [".tif", ".tiff"]:
            pass
        elif self.path.suffix in [".nc"]:
            xda = xarray.open_dataarray(
                self.path.absolute(), decode_coords="all"
            )
            resampled = xda.resample(time=self.time.freq).mean()
            xda_now = resampled.sel(
                time=self.time.period.to_timestamp("S"),
                method=selecting_method,
            )
            matched_xda = self.geo.clip_match(xda_now)
            return matched_xda.to_numpy()

    pass



#!/usr/bin/env python 3.11.0
# -*-coding:utf-8 -*-
# @Author  : Shuang (Twist) Song
# @Contact   : SongshGeo@gmail.com
# GitHub   : https://github.com/SongshGeo
# Website: https://cv.songshgeo.com/
import logging
from typing import Any, Iterable, List

import numpy as np

logger = logging.getLogger(__name__)


def make_list(element, keep_none=False):
    """Turns element into a list of itself
    if it is not of type list or tuple."""

    if element is None and not keep_none:
        element = []  # Convert none to empty list
    if not isinstance(element, (list, tuple, set, np.ndarray)):
        element = [element]
    elif isinstance(element, (tuple, set)):
        element = list(element)

    return element


def norm_choice(
    a: Iterable[any], size: int = None, p=None, replace=False
) -> Any:
    """
    A more robust random chooser.

    Args:
        a (Iterable[any]): An iterable instance to choose.
        size (int, optional): size of the choice. Defaults to None.
        p (_type_, optional): probability. Defaults to None.
        replace (bool, optional): _description_. Defaults to False.

    Returns:
        Any: _description_
    """
    p = np.array(make_list(p))
    possible_entries = (p > 0).sum()
    if possible_entries == 0:
        p = None
    else:
        p = np.where(p > 0, p, 0.0)
        p /= p.sum()
    # solve probability
    if p is None or replace is True:
        return np.random.choice(a, size=size, p=p, replace=replace)
    else:
        # 如果有概率的实体 少于要选择的实体，优先返回
        if size is None:
            return np.random.choice(a, p=p, replace=False)
        elif possible_entries < size:
            bounds = a[p > 0]
            rand = np.random.choice(
                a[p == 0], size=(size - possible_entries), replace=False
            )
            selected = np.concatenate([bounds, rand])
        else:
            selected = np.random.choice(a, p=p, size=size, replace=False)
    return selected


def unique_list(*args):
    unique = set()
    for lst in args:
        if not isinstance(lst, Iterable):
            raise TypeError("unique_list can only convert iterable to list")
        unique = unique | set(lst)
    return list(unique)


def iter_func(elements: str) -> callable:
    """
    A decorator broadcasting function to all elements if available.

    elements:
        elements (str): attribute name where object store iterable elements.
        All element in this iterable object will call the decorated function.
    """

    def broadcast(func: callable) -> callable:
        def broadcast_func(self, *args, **kwargs):
            result = func(self, *args, **kwargs)
            if not hasattr(self, elements):
                return result
            for element in getattr(self, elements):
                getattr(element, func.__name__)(*args, **kwargs)
            return result

        return broadcast_func

    return broadcast


def opfunc_using_attr(attr: str, binary_op: bool) -> callable:
    """
    wrap a operation method of a custom class, for using its 'attribution' in calculation.

    Args:
        attr (str): which attribute of the class to use in calculation.
        binary_op (bool): if the wrapping operation method is binary or unary.

    Returns:
        callable: a wrapped bound method.

    Examples:
        class Calculation:
            def __init__(self):
                self.attr = 1

        cal = Calculation()

        cal + 1
        >>> TypeError: 'unsupported operand type(s)' ...

        # wraping the '__add__' method.
        opname = '__add__'
        wrapper = opfunc_using_attr('attr', binary_op=True)
        wrapped_func = wrapper(getattr(1, opname))
        setattr(cal.__class__, opname, wrapped_func)

        cal + 1 == 2
        >>> 2
    """

    # wraping binary operation functions.
    def wrap_binary_opfunc(func):
        def new_func(self, other):
            opfunc = getattr(getattr(self, attr), func.__name__)
            if issubclass(other.__class__, self.__class__):
                return opfunc(getattr(other, attr))
            else:
                return opfunc(other)

        return new_func

    # wraping unary operation functions.
    def wrap_unary_opfunc(func):
        def new_func(self):
            opfunc = getattr(getattr(self, attr), func.__name__)
            return opfunc()

        return new_func

    if binary_op:
        return wrap_binary_opfunc
    else:
        return wrap_unary_opfunc


# https://zhuanlan.zhihu.com/p/515284250
_binary_opfunc = (
    "__add__",
    "__sub__",
    "__mul__",
    "__truediv__",
    "__floordiv__",
    "__mod__",
    "__divmod__",
    "__pow__",
    "__radd__",
    "__rsub__",
    "__rmul__",
    "__rtruediv__",
    "__rfloordiv__",
    "__rmod__",
    "__rdivmod__",
    "__rpow__",
    "__lshift__",
    "__rshift__",
    "__and__",
    "__xor__",
    "__or__",
    "__contains__",
)

_unary_opfunc = (
    "__neg__",
    "__pos__",
    "__abs__",
    "__invert__",
)


def wrap_opfunc_to(obj: object, attr: str) -> List[str]:
    """
    Overwrite an object's all operation functions, using its attribute in future's calculations.

    Args:
        obj (object): any object to be wrapped.
        attr (str): object's attribute name.

    Returns:
        List[str]: successful wrapped operation functions' name.

    Examples:
        class MyClass(object):
            def __init__(self, attr: any):
                self.attr: any = attr
        int_obj = MyClass(1)
        str_obj = MyClass('test')
        wrap_opfunc_to
    """

    def wrap_opfunc(binary_op):
        func = getattr(getattr(obj, attr), opname, None)
        if func is None:
            return False
        else:
            wrapper = opfunc_using_attr(attr, binary_op=binary_op)
            setattr(obj.__class__, opname, wrapper(func))
            return True

    successful_wrapped = []
    for opname in _binary_opfunc:
        wrapped = wrap_opfunc(True)
        if wrapped:
            successful_wrapped.append(opname.strip("_"))
    for opname in _unary_opfunc:
        wrapped = wrap_opfunc(False)
        if wrapped:
            successful_wrapped.append(opname.strip("_"))
    return successful_wrapped



#!/usr/bin/env python 3.11.0
# -*-coding:utf-8 -*-
# @Author  : Shuang Song
# @Contact   : SongshGeo@gmail.com
# GitHub   : https://github.com/SongshGeo
# Website: https://cv.songshgeo.com/
import copy
from collections.abc import Iterable
from functools import cached_property
from typing import Optional

import numpy as np
import rioxarray
import xarray as xr
from scipy import ndimage

from .bases import Creation
from .geo import Geo

# from src.abm_chans.visualization import PatchViz


# TODO use a decorator function instead of recalc


def get_buffer(array, buffer=1, neighbors=4):
    if neighbors == 4:
        connectivity = 1
    elif neighbors == 8:
        connectivity = 2
    struct = ndimage.generate_binary_structure(2, connectivity)
    result = array.copy()
    i = 1
    while i <= buffer:
        result = ndimage.binary_dilation(result, structure=struct).astype(
            array.dtype
        )
        i += 1
    return result


def update_array(
    array,
    value: "np.ndarray|str|bool|float",
    mask: np.ndarray = None,
):
    """
    Update a groundwater 2-d array's attributions.

    Arguments:
        value -- Change to the new value. Can be a new array (shape is same as the masked matrix' shape.) or a single value.
        mask -- Which element to change, default {None}, then all elements will be set.

    Returns:
        Changed mask, where have been changed.
    """
    if mask is None:
        mask = np.ones(array.shape, dtype=bool)
    if isinstance(value, Iterable):
        if value.shape == array.shape:
            array[mask] = value[mask]
        elif len(value) == mask.sum():
            array[mask] = value
        else:
            raise TypeError(f"Not support {type(value)}")
    else:
        array[mask] = value
    return array


class ArrayOperation:
    def __init__(self, array):
        self.array = array

    def buffer(self, buffer=1, neighbors=4):
        return get_buffer(self.array, buffer=buffer, neighbors=neighbors)

    def unique(self) -> np.ndarray:
        return np.unique(self.array)

    def sort(self, *args, **kwargs) -> np.ndarray:
        axis = kwargs.pop("axis", None)
        return np.sort(self.array, axis=axis, *args, **kwargs)

    def where(self):
        return list(zip(*np.where(self.array)))


class GeoXarray:
    def __init__(self, xda: xr.DataArray, geo: Geo):
        self._obj: xr.DataArray = self.setup_spatial(xda, geo)
        self._geo: Geo = geo

    def setup_spatial(self, xda, geo: Geo) -> xr.DataArray:
        xda = xda.rio.write_crs(geo.crs)
        xda = xda.rio.set_spatial_dims(x_dim=geo.dims[0], y_dim=geo.dims[0])
        xda = xda.rio.set_nodata(geo.nodata)
        xda = xda.rio.write_coordinate_system()
        return xda

    @property
    def resolution(self):
        affine = self._obj.rio.transform()
        size_x = abs(affine[0])
        size_y = abs(affine[4])
        return size_x, size_y

    @cached_property
    def area(self):
        x, y = self.resolution
        area = x * y
        self._area = area
        return area

    def resampling(self, factor):
        from rasterio.enums import Resampling

        xds = self._obj.transpose("lat", "lon")
        new_width = xds.rio.width * factor
        new_height = xds.rio.height * factor
        scaled = xds.rio.reproject(
            xds.rio.crs,
            shape=(new_height, new_width),
            resampling=Resampling.bilinear,
        )
        return scaled


class Patch(np.ndarray, Creation):
    def __new__(
        cls,
        array: np.ndarray,
        name: Optional[str] = None,
        father: Optional[object] = None,
        xarray: Optional[bool] = False,
    ):
        obj = np.asarray(array).view(cls)
        obj._farther = father
        obj._name = name
        obj._check_valid()
        obj._xda: xr.DataArray = None
        obj._rio: GeoXarray = None
        if xarray:
            obj.to_xarray(recalc=True)
        return obj

    @property
    def arr(self) -> ArrayOperation:
        return ArrayOperation(self.view(np.ndarray))

    # @property
    # def plot(self) -> PatchViz:
    #     xarray = self.xda
    #     return PatchViz(xarray)

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def father(self):
        return self._farther

    @property
    def xda(self) -> xr.DataArray:
        return self._xda

    @property
    def rio(self) -> GeoXarray:
        if self._rio is None:
            return self.to_spatial(True)
        else:
            return self._rio

    @property
    def cell_area(self):
        # TODO: refactor this into global variable
        return self.rio.area()

    def _check_valid(self) -> bool:
        if len(self.shape) != 2:
            raise ValueError("Invalid shape.")
        return True

    def __array_finalize__(self, obj):
        if obj is None:
            return
        self.name = getattr(obj, "name", None)

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        # this method is called whenever you use a ufunc
        # i.e., symbols like +, -, *, / ...
        # https://stackoverflow.com/questions/51520630/subclassing-numpy-array-propagate-attributes
        f = {
            "reduce": ufunc.reduce,
            "accumulate": ufunc.accumulate,
            "reduceat": ufunc.reduceat,
            "outer": ufunc.outer,
            "at": ufunc.at,
            "__call__": ufunc,
        }
        output = f[method](
            *(
                i.view(np.ndarray) if isinstance(i, Patch) else i
                for i in inputs
            ),
            **kwargs,
        )
        # output.__dict__ = self.__dict__  # carry forward attributes
        return output

    def update(
        self,
        value: "np.ndarray|str|bool|float",
        mask: np.ndarray = None,
    ):
        """
            Update a groundwater 2-d array's attributions.

            Arguments:
                value -- Change to the new value. Can be a new array (shape is same as the masked matrix' shape.) or a single value.
                mask -- Which element to change, default {None}, then all elements will be set.

        Returns:
            Changed mask, where have been changed.
        """
        update_array(self, value, mask)
        if self.xda is not None:
            update_array(self.xda.data, value, mask)

    def to_xarray(self, recalc: bool = True) -> xr.DataArray:
        if recalc is False:
            return self._xda
        if self.father is None:
            raise ValueError("father of patch must be specified")
        xda = xr.DataArray(
            data=self,
            name=self.name,
            coords=self.father.geo.coords,
            attrs=self.father.attrs,
        ).where(self.father.accessible)
        self._xda = xda
        return xda

    def to_spatial(self, recalc: bool = False) -> None:
        if self.xda is None:
            # raise AB_EGMpyError("patch", wrong="convert", condition="")
            raise ValueError("to_spatial requires patch")
        if recalc is True:
            self._rio = GeoXarray(self.xda, self.father)
        return self.rio



#!/usr/bin/env python 3.11.0
# -*-coding:utf-8 -*-
# @Author  : Shuang Song
# @Contact   : SongshGeo@gmail.com
# GitHub   : https://github.com/SongshGeo
# Website: https://cv.songshgeo.com/

from typing import Callable, Dict, Optional, TypeAlias, Union

import networkx as nx
from agentpy import AttrDict

from abses.actor import Actor
from abses.bases import Creation

from .container import AgentsContainer
from .modules import CompositeModule, Module
from .sequences import ActorsList, Selection

Actors: TypeAlias = Union[ActorsList, Selection, Actor]
Trigger: TypeAlias = Union[str, Callable]


class HumanModule(Module, Creation):
    def __init__(self, model, name=None):
        Module.__init__(self, model, name)
        Creation.__init__(self)
        self._agents = AgentsContainer(model)
        self._collections: Dict[str, Selection] = AttrDict()
        self._rules: Dict[str, Trigger] = AttrDict()

    def __getattr__(self, name):
        if name[0] == "_":
            return super().__getattr__(name)
        elif name in self._collections:
            selection = self._collections[name]
            return self.actors.select(selection)
        else:
            return super().__getattr__(name)

    @property
    def agents(self) -> AgentsContainer:
        return self._agents

    @property
    def actors(self) -> ActorsList:
        return self.agents.to_list()

    def define(self, name: str, selection: Selection) -> ActorsList:
        selected = self.actors.select(selection)
        self._collections[name] = selection
        return selected

    # def rule(self, actors: Actors, when: Selection, then: Trigger, name: Optional[str] = None):
    #     if name is None:
    #         pass
    #     self.define(name=name)
    #     actors_to_trigger = actors.select(when)
    #     results = actors_to_trigger.trigger(then)
    #     return actors_to_trigger, results

    def arena(self, actor_A: Actors, actor_B: Actors, interaction: Trigger):
        actor_A.trigger(interaction, actor_B)
        actor_B.trigger(interaction, actor_A)

    def require(self, attr: str) -> object:
        return self.mediator.transfer_require(self, attr)


class BaseHuman(CompositeModule, HumanModule):
    def __init__(self, model, name="human"):
        HumanModule.__init__(self, model, name)
        CompositeModule.__init__(self, model, name=name)


#     def mock(self, agents, attrs, how="attr"):
#         tutors = self.to_agents(agents.tutor.now)
#         for attr in make_list(attrs):
#             values = tutors.array(attr, how)
#             agents.update(attr, values)


# def skip_if_close(func):
#     def skip_module_method(self, *args, **kwargs):
#         if self.opening:
#             func(self, *args, **kwargs)
#         else:
#             if self.log_flag:
#                 self.logger.warning(f"{self}.")

#     return skip_module_method



#!/usr/bin/env python 3.11.0
# -*-coding:utf-8 -*-
# @Author  : Shuang Song
# @Contact   : SongshGeo@gmail.com
# GitHub   : https://github.com/SongshGeo
# Website: https://cv.songshgeo.com/
from __future__ import annotations

from collections.abc import Iterable
from typing import List, Optional, Set, Union

import numpy as np
from agentpy import Model
from prettytable import PrettyTable

from .bases import Creator
from .components import Component, MainComponent, iter_func
from .objects import BaseObj
from .tools.func import make_list


class Module(Component, BaseObj):
    def __init__(self, model: Model, name: Optional[str] = None):
        Component.__init__(self, name=name)
        BaseObj.__init__(self, model, observer=True, name=name)
        self._model: Model = model
        self._open: bool = True

    def __repr__(self) -> str:
        if self.opening:
            flag = "open"
        else:
            flag = "closed"
        return f"<{self.name}: {flag}>"

    @property
    def opening(self) -> bool:
        return self._open

    def _after_parsing(self):
        self.switch_open_to(self.params.pop("open", None))
        self.recording = self.params.pop("record", [])
        self.reporting = self.recording  # recording vars also reported
        self.reporting = self.params.pop("report", [])

    @iter_func("modules")
    def report_vars(self):
        for var in self._reporting:
            value = getattr(self, var)
            self.model.report(var, value)

    @iter_func("modules")
    def switch_open_to(self, _open: Optional[bool] = None) -> bool:
        """#TODO 思考和说明模块关闭后有何不同"""
        if _open is None:
            return False
        elif not isinstance(_open, bool):
            raise TypeError("Accept boolean parameters")
        else:
            if self._open is not _open:
                self.logger.info(f"{self} switch 'open' to {_open}.")
                self._open = _open
            else:
                pass
        return self._open


# Composite
class CompositeModule(Module, MainComponent, Creator):
    def __init__(self, model: Model, name: str = None) -> None:
        MainComponent.__init__(self, name=name)
        Creator.__init__(self)
        Module.__init__(self, model, name=name)
        self._modules: List[Module] = []

    @property
    def modules(self) -> List[Module]:
        return self._modules

    def summary_modules(self) -> PrettyTable:
        table = PrettyTable()
        table.field_names = ["Name", "Opening", "Params"]
        for module in self.modules:
            table.add_row([module.name, module.opening, len(module.params)])
        return table

    def create_module(self, module_class: Module, *args, **kwargs) -> Module:
        module = module_class(model=self.model, *args, **kwargs)
        if not issubclass(module.__class__, Module):
            raise TypeError("Must inherited from a module.")
        setattr(self, module.name, module)  # register as module
        self.add_creation(module)  # register as creation
        self.modules.append(module)  # register as module
        self.notify()  # update attributes
        return module



#!/usr/bin/env python 3.11.0
# -*-coding:utf-8 -*-
# @Author  : Shuang (Twist) Song
# @Contact   : SongshGeo@gmail.com
# GitHub   : https://github.com/SongshGeo
# Website: https://cv.songshgeo.com/

import numpy as np
from matplotlib.path import Path
from scipy.spatial import Voronoi


def points_to_polygons(points: np.ndarray) -> np.ndarray:
    if not (points.ndim == 2) & (points.shape[1] == 2):
        raise ValueError
    # points[:, ::-1] reversed x, y of a point
    vor = Voronoi(points[:, ::-1])
    regions, vertices = voronoi_finite_polygons_2d(vor)
    polygons = []
    for reg in regions:
        polygon = vertices[reg]
        polygons.append(polygon)
    return polygons


def voronoi_finite_polygons_2d(vor, radius=None):
    """Reconstruct infinite Voronoi regions in a
    2D diagram to finite regions.
    Source:
    [https://stackoverflow.com/a/20678647/1595060](https://stackoverflow.com/a/20678647/1595060)
    The `voronoi_finite_polygons_2d()` function returns a list of regions and a list of vertices. Every region is a list of vertex indices. The coordinates of all vertices are stored in vertices. From these structures, we can create a list of cells. Every cell represents a polygon as an array of vertex coordinates. We also use the to_pixels() method of the smopy.Map instance. This function converts latitude and longitude geographical coordinates to pixels in the image.
    """
    if vor.points.shape[1] != 2:
        raise ValueError("Requires 2D input")
    new_regions = []
    new_vertices = vor.vertices.tolist()
    center = vor.points.mean(axis=0)
    if radius is None:
        radius = vor.points.ptp().max()
    # Construct a map containing all ridges for a
    # given point
    all_ridges = {}
    for (p1, p2), (v1, v2) in zip(vor.ridge_points, vor.ridge_vertices):
        all_ridges.setdefault(p1, []).append((p2, v1, v2))
        all_ridges.setdefault(p2, []).append((p1, v1, v2))
    # Reconstruct infinite regions
    for p1, region in enumerate(vor.point_region):
        vertices = vor.regions[region]
        if all(v >= 0 for v in vertices):
            # finite region
            new_regions.append(vertices)
            continue
        # reconstruct a non-finite region
        ridges = all_ridges[p1]
        new_region = [v for v in vertices if v >= 0]
        for p2, v1, v2 in ridges:
            if v2 < 0:
                v1, v2 = v2, v1
            if v1 >= 0:
                # finite ridge: already in the region
                continue
            # Compute the missing endpoint of an
            # infinite ridge
            t = vor.points[p2] - vor.points[p1]  # tangent
            t /= np.linalg.norm(t)
            n = np.array([-t[1], t[0]])  # normal
            midpoint = vor.points[[p1, p2]].mean(axis=0)
            direction = np.sign(np.dot(midpoint - center, n)) * n
            far_point = vor.vertices[v2] + direction * radius
            new_region.append(len(new_vertices))
            new_vertices.append(far_point.tolist())
        # Sort region counterclockwise.
        vs = np.asarray([new_vertices[v] for v in new_region])
        c = vs.mean(axis=0)
        angles = np.arctan2(vs[:, 1] - c[1], vs[:, 0] - c[0])
        new_region = np.array(new_region)[np.argsort(angles)]
        new_regions.append(new_region.tolist())
    return new_regions, np.asarray(new_vertices)


def polygon_to_mask(polygon, shape):
    height = shape[0]
    width = shape[1]
    poly_path = Path(polygon)
    x, y = np.mgrid[:width, :height]
    coors = np.hstack((x.reshape(-1, 1), y.reshape(-1, 1)))
    mask = poly_path.contains_points(coors)
    return mask.reshape(width, height).T



#!/usr/bin/env python 3.11.0
# -*-coding:utf-8 -*-
# @Author  : Shuang (Twist) Song
# @Contact   : SongshGeo@gmail.com
# GitHub   : https://github.com/SongshGeo
# Website: https://cv.songshgeo.com/
from __future__ import annotations

import logging
from abc import ABCMeta, abstractmethod
from typing import Iterable, List, Optional, Set, Union

from agentpy.objects import make_list

logger = logging.getLogger(__name__)


class Creator:
    def __init__(self):
        self.created: List[Creation] = []
        self._inheritance: List[str] = []

    def _check_creation(self, creation: Creation):
        if not issubclass(creation.__class__, Creation):
            raise TypeError(
                f"Only creation can be added, input {creation.__class__}."
            )

    @property
    def inheritance(self) -> list:
        return self._inheritance

    @inheritance.setter
    def inheritance(self, attrs: "List[str]|str") -> None:
        self.inheritance.extend(make_list(attrs))
        self.notify()

    def add_creation(self, obj: Creation):
        self._check_creation(creation=obj)
        self.created.append(obj)
        setattr(obj, "_creator", self)
        self.notify()

    def notify(self):
        for obj in self.created:
            obj.refresh(self)

    def remove_creation(self, obj: Creation):
        self._check_creation(creation=obj)
        self.created.remove(obj)


class Creation(metaclass=ABCMeta):
    @property
    def creator(self):
        return self._creator

    def refresh(self, creator: Creator):
        for attr in creator.inheritance:
            value = creator.__getattribute__(attr)
            self.__setattr__(attr, value)


class Notice:
    def __init__(self, observer: Optional[Observer] = None):
        self.observers: Set[Observer] = set()
        self._glob_vars: List[str] = []
        if observer is not None:
            self.attach(observer)

    def __repr__(self) -> str:
        return f"<{len(self.observers)} are observing: {self.glob_vars}>"

    @property
    def glob_vars(self) -> List[str]:
        return self._glob_vars

    @glob_vars.setter
    def glob_vars(self, value: "str|List[str]"):
        self._glob_vars.extend(make_list(value))
        self.notify()

    def attach(self, observer: Observer):
        self.observers.add(observer)
        observer.notification(self)

    def detach(self, observer: Observer):
        self.observers.remove(observer)

    def notify(self, *args: str):
        if len(args) > 0:
            for attr in args:
                val = getattr(self, attr)
                for observer in self.observers:
                    setattr(observer, attr, val)
            logger.info(f"send {args} to {self.observers}")
        for observer in self.observers:
            observer.notification(self)


class Observer(metaclass=ABCMeta):
    def notification(self, notice: Notice):
        self.glob_vars = notice.glob_vars
        for var in self.glob_vars:
            value = getattr(notice, var)
            setattr(self, var, value)

    @property
    def glob_vars(self) -> List[str]:
        return self._glob_vars

    @glob_vars.setter
    def glob_vars(self, value: Union[Iterable[str], str]) -> None:
        if not hasattr(self, "_glob_vars"):
            self._glob_vars = []
        self._glob_vars.extend(make_list(value))


class Mediator:
    @abstractmethod
    def transfer_event(self, sender: object, event: str):
        pass

    @abstractmethod
    def transfer_parsing(self, sender: object, params: dict):
        pass

    @abstractmethod
    def transfer_request(self, sender: object, request: str):
        return request



#!/usr/bin/env python 3.11.0
# -*-coding:utf-8 -*-
# @Author  : Shuang Song
# @Contact   : SongshGeo@gmail.com
# GitHub   : https://github.com/SongshGeo
# Website: https://cv.songshgeo.com/
from __future__ import annotations

import logging
from numbers import Number
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    Self,
    Tuple,
    TypeAlias,
    Union,
    overload,
)

import networkx as nx
import numpy as np
from agentpy import AgentSet, AttrDict

from .objects import BaseObj
from .patch import Patch

if TYPE_CHECKING:
    from .main import MainMediator
    from .sequences import ActorsList

Selection: TypeAlias = Union[str, Iterable[bool]]
Trigger: TypeAlias = Union[Callable, str]

logger = logging.getLogger("__name__")


def parsing_string_selection(selection: str) -> Dict[str, Any]:
    selection_dict = dict()
    if "==" not in selection:
        return {"breed": selection}
    else:
        expressions = selection.split(",")
        for exp in expressions:
            left, right = tuple(exp.split("=="))
            selection_dict[left.strip(" ")] = right.strip(" ")
    return selection_dict


def perception(func):
    @property
    def wrapper(self: Actor, *args, **kwargs):
        return func(self, *args, **kwargs)

    return wrapper


def link_to(func):
    # TODO and links, which can be searched through networkx
    @property
    def wrapper(self: Actor, *args, **kwargs):
        return func(self, *args, **kwargs)

    return wrapper


def check_rule(loop: bool = False) -> Callable:
    def f(func: Callable) -> Callable:
        def wrapper(self: Actor, *args, **kwargs):
            triggered = self._check_rules("now")
            if loop is True:
                while len(triggered) > 0:
                    triggered = self._check_rules("now")
            return func(self, *args, **kwargs)

        return wrapper

    return f


class Actor(BaseObj):
    # when checking the rules
    _freq_levels = {"now": 0, "update": 1, "move": 2, "any": 3}

    def __init__(
        self,
        model,
        observer: bool = True,
        name: Optional[str] = None,
        **kwargs,
    ):
        BaseObj.__init__(self, model, observer=observer, name=name)
        self._on_earth: bool = False
        self._pos: Tuple[int, int] = None
        self._relationships: Dict[str, ActorsList] = AttrDict()
        self._ownerships: Dict[str, Patch] = AttrDict()
        self._rules: Dict[str, Dict[str, Any]] = dict()
        self.mediator: MainMediator = self.model.mediator
        self.setup(**kwargs)

    def __setattr__(self, name, value):
        super().__setattr__(name, value)
        if name[0] != "_" and hasattr(self, "_rules"):
            self._check_rules(check_when="any")

    @classmethod
    @property
    def breed(cls) -> str:
        return cls.__name__

    @property
    def population(self):
        return self.model.agents[self.breed]

    @property
    def on_earth(self) -> bool:
        return self._on_earth

    @property
    def pos(self) -> Tuple[int, int]:
        return self._pos

    @property
    def here(self) -> ActorsList:
        if self.on_earth is True:
            return self.neighbors(0)
        else:
            return None

    def _freq_level(self, level: str) -> int:
        code = self._freq_levels.get(level, None)
        if code is None:
            keys = tuple(self._freq_levels.keys())
            raise KeyError(f"freq level {level} is not available in {keys}")
        return code

    def _check_rules(self, check_when: str) -> List[str]:
        triggered_rules = []
        for name in list(self._rules.keys()):
            rule = self._rules[name]
            parameters = rule.get("params", {})
            if rule.frequency < self._freq_level(check_when):
                continue
            if self.selecting(rule.when) is True:
                triggered_rules.append(name)
                # check if is a disposable rule
                if rule.disposable is True:
                    self.logger.debug(
                        f"Rule '{name}' applied on '{self}' in {self.time}."
                    )
                    del self._rules[name]
                self.__getattr__(rule.then)(**parameters)
        # delete disposable rules
        return triggered_rules

    def request(
        self,
        request: str,
        header: Dict[str, Any],
        receiver: Optional[str] = None,
    ) -> Any:
        if receiver is None:
            response = self.mediator.transfer_request(self, request)
        elif receiver in ["nature", "human"]:
            results = self.mediator.trigger_functions(
                users=receiver, func_name=request, **header
            )
            response = results.__getattribute__(receiver)
        else:
            raise ValueError(f"Unknown transfer request {receiver}")
        return response

    # def request(self, request: str, header=None, receiver=None) -> Any:
    #     # header.update({'how': 'GET'})
    #     return self.request(request, header, receiver)

    # def post(self, request: str, value: Any, header=None, receiver=None):
    #     header.update({'how': 'POST', request: value})
    #     return self._request(request, header, receiver)

    def selecting(self, selection: Union[str, Dict[str, Any]]) -> bool:
        if isinstance(selection, str):
            selection = parsing_string_selection(selection)
        results = []
        for k, v in selection.items():
            attr = getattr(self, k, None)
            if attr is None:
                results.append(False)
            elif attr == v or str(attr) == v:
                results.append(True)
            else:
                results.append(False)
        return all(results)

    def rule(
        self,
        when: Selection,
        then: Trigger,
        name: Optional[str] = None,
        frequency: str = "any",
        disposable: bool = False,
        check_now: Optional[bool] = True,
        **kwargs,
    ):
        if name is None:
            name = f"rule ({len(self._rules) + 1})"
        self._rules[name] = AttrDict(
            {
                "when": when,
                "then": then,
                "params": kwargs,
                "disposable": disposable,
                "frequency": self._freq_level(frequency),
            }
        )
        if check_now is True:
            self._check_rules("now")

    def die(self):
        self.model.agents.remove(self)

    def neighbors(
        self,
        distance: int = 1,
        approach: int = 4,
        selection: Selection = None,
        exclude: bool = True,
    ):
        # The area around within a certain distance.
        header = {
            "pos": self.pos,
            "distance": distance,
            "selection": selection,
            "approach": approach,
        }
        agents = self.request(
            request="neighbors", header=header, receiver="nature"
        )
        if exclude is True:
            agents.remove(self)
        return agents

    def settle_down(self, position: Optional[Tuple[int, int]]) -> bool:
        header = {"actor": self, "position": position}
        self.request("actor_to", header=header, receiver="nature")
        self._pos = position
        self._on_earth = True

    def move(self, pos: Optional[Tuple[int, int]] = None):
        if self.on_earth is False:
            raise ValueError(f"Position of {self} is not set.")
        if pos is None:
            pos = self.request(
                "random_positions", header={"k": 1}, receiver="nature"
            )[0]
        self.settle_down(pos)

    def link_to(self, name: str, other: Union[Self, Iterable[Self], Callable]):
        pass

    def loc(self, request: str, **kwargs):
        header = {"sender": self, "position": self.pos}
        response = self.request(request, header, **kwargs)
        return response[self.pos]
        # elif len(patch_obj.shape) == 3:
        #     return patch_obj[:, self.pos[0], self.pos[1]]

    def alter_nature(self, patch: str, value: "int|float|bool") -> Patch:
        patch = self.require(patch)
        patch[self.pos] = value
        return patch

    # def find_tutor(self, others, metric, how="best"):
    #     better = others.better(metric, than=self)
    #     if len(better) == 0:
    #         tutor = self
    #     else:
    #         if how == "best":
    #             better = others.better(metric, than=None)
    #             weights = None
    #         elif how == "diff":
    #             diff = better.diff(metric, self)
    #             weights = diff
    #         # elif how == "fermi":
    #         #     diff = better.diff(metric, self)
    #         #     weights = fermi_ruler(diff)
    #         else:
    #             raise ValueError(f"Unknown evolve strategy {how}.")
    #         tutor = better.random_choose(p=weights)
    #     self.tutor.update(tutor)
    #     return tutor


# class LandOwner(Actor):
#     def setup(self):
#         super().setup()
#         self._owned_land: Patch = None
#         self._area: Number = None
#         self.land_neighbors = ActorsList(self.model)

#     @property
#     def area(self):
#         return self._area

#     @property
#     def owned_land(self):
#         return self._owned_land

#     @owned_land.setter
#     def owned_land(self, value: "np.ndarray[bool]|bool"):
#         name = f"land_of_{self.breed}<{self.id}>"
#         self._owned_land = self.model.nature.create_patch(value, name=name)
#         self._area = self.owned_land.sum() * self.owned_land.cell_area

#     @property
#     def __iter_lands__(self):
#         return self.owned_land.cells

#     def residents(self, breed: str = None):
#         return self.model.nature.lookup_agents(self.owned_land, breed=breed)

#     def mine(self, patch: str, apply: callable = None) -> np.ndarray:
#         mine = self.require(patch)[self.owned_land]
#         if apply is not None:
#             return apply(mine)
#         else:
#             return mine

#     def add_connections(self, graph: nx.Graph, agents: Iterable[Actor]):
#         graph.add_edges_from([(self, agent) for agent in agents])

#     def alter_nature(
#         self, patch: str, value: "int|float|bool|Iterable", land=False
#     ) -> None:
#         if land:
#             mask = self.owned_land
#             patch = self.require(patch)
#             patch.update(value, mask=mask)
#         else:
#             super().alter_nature(patch, value)

#     def find_land_neighbors(self, ownerships, graph: nx.Graph = None) -> None:
#         owned = self.owned_land
#         mask = owned.arr.buffer() & ~owned
#         neighbors_id = np.unique(ownerships[mask])
#         neighbors = self.population.ids(neighbors_id)
#         self.land_neighbors.extend(neighbors)
#         self.add_connections(graph, neighbors)

#     def find_tutor(self, metric, others=None, how="best"):
#         if others is None:
#             others = self.land_neighbors
#         return super().find_tutor(others, metric, how)



#!/usr/bin/env python 3.11.0
# -*-coding:utf-8 -*-
# @Author  : Shuang Song
# @Contact   : SongshGeo@gmail.com
# GitHub   : https://github.com/SongshGeo
# Website: https://cv.songshgeo.com/

from functools import cached_property
from typing import Optional, Tuple

import numpy as np
import xarray

from .bases import Creator
from .boundary import Boundaries, simple_boundary_from
from .geo import Geo
from .patch import Patch

# def generate_boundary(self, settings: Optional[dict] = None) -> Boundaries:
#     # resolution = settings.pop("resolution", 1)
#     boundary = simple_boundary_from(settings)
#     self.shape = boundary.shape
#     # self.coords = self.setup_coords(width, height, resolution)
#     self.mask = ~boundary.interior
#     self.boundary = boundary
#     self.notify()
#     return boundary



#!/usr/bin/env python
# -*-coding:utf-8 -*-
# Created date: 2022/1/14
# @Author  : Shuang Song
# @Contact   : SongshGeo@gmail.com
# GitHub   : https://github.com/SongshGeo
# Website: https://cv.songshgeo.com/

import logging

import numpy as np

# from .tools.plot_heatmap import custom_cmap_heatmap, matrix_heatmap

logger = logging.getLogger(__name__)


class Boundaries:

    BOUNDARIES = ["outside", "interior", "boundaries", "special_boundaries"]

    def __init__(self, shape):
        self.shape = shape
        self._special_boundaries = {}  # Named boundaries' dict.
        self._boundaries = np.zeros(shape, dtype=bool)
        self._outside = np.zeros(shape, dtype=bool)
        self._interior = np.ones(shape, dtype=bool)
        self._update_all()

    @property
    def boundaries(self):
        """All boundaries."""
        return self._boundaries

    @property
    def special_boundaries(self):
        """Named special boundaries."""
        return self._special_boundaries

    @property
    def outside(self):
        """Outside the boundary."""
        return self._outside

    @property
    def interior(self):
        return self._interior

    def _check_shape(self, ndarray):
        """Check if the input ndarray's shape can fit in."""
        if ndarray.shape != self.shape:
            raise ValueError("shape must be 2d")
        else:
            return True

    def _update_all(self):
        """Update all boundaries"""
        changed_boundaries = self._update_boundaries()
        changed_interior = self._update_interior()
        changed = changed_boundaries == changed_interior
        perfect_change = changed.all()
        if not perfect_change:
            logger.warning("Update not perfect!")
        return changed

    def _update_interior(self):
        """Update interior mask by the boundaries mask."""
        new_interior = ~(self.boundaries | self.outside)
        changed = self.interior != new_interior
        self.reset_boundary("interior", new_interior)
        return changed

    def _update_boundaries(self):
        """Update all boundary according to all types of boundary."""
        new_boundary = self.boundaries
        for bc_type in self.special_boundaries:
            new_boundary = new_boundary | self.special_boundaries[bc_type]
        changed = self.boundaries != new_boundary
        self.reset_boundary("boundaries", new_boundary)
        return changed

    def init_simple_boundaries(self, shape: tuple, settings: dict = None):
        outsides = settings.pop("outsides", 0)
        border = settings.pop("border", 1)
        boundary = generate_simple_boundaries(
            shape=shape,
            boundary=self,
            outsides=outsides,
            border=border,
        )
        return boundary

    def change_special_boundary(self, name, patch):
        """Update a special boundary."""
        self._check_shape(patch)
        self._special_boundaries[name] = patch
        self._update_all()

    def reset_boundary(self, boundary_type: str, ndarray: np.ndarray):
        """Update any type of boundary (default or named)."""
        self._check_shape(ndarray)
        if hasattr(self, boundary_type):
            setattr(self, f"_{boundary_type}", ndarray)
        elif boundary_type in self.special_boundaries:
            self.change_special_boundary(boundary_type, ndarray)
        else:
            raise KeyError(f"No boundary type named {boundary_type}.")
        return ndarray

    # def plot_boundaries(self, bc_type="all", annotate_names=False, **kwargs):
    #     default_bc_type = ("boundaries", "interior", "outside")
    #     colors_map = {
    #         "Interior": "gray",
    #         "Boundary": "black",
    #         "Outside": "lightgray",
    #     }
    #     if annotate_names:
    #         annot = np.ones(self.shape, dtype=str)
    #         annot.fill("")
    #         for name, named_matrix in self.special_boundaries.items():
    #             annot[named_matrix] = name.capitalize()[0]
    #         kwargs["annot"] = annot
    #         kwargs["fmt"] = ""

    #     if bc_type == "all":
    #         outside = self.outside
    #         boundaries = self.boundaries
    #         interior = self.interior
    #         matrix = 0 * interior + 0.5 * boundaries + outside
    #         ax = custom_cmap_heatmap(
    #             matrix, colors_map, vmin=0, vmax=1, **kwargs
    #         )

    #     elif bc_type in default_bc_type:
    #         matrix = self.__getattribute__(bc_type)
    #         ax = matrix_heatmap(
    #             matrix, how="mask", title=f"{bc_type}", **kwargs
    #         )

    #     else:
    #         raise KeyError(f"{bc_type} not in {default_bc_type}.")
    #     return ax


def set_value_to_border(ndarray, thickness, value):
    """
    Set a 2d array's border to a specified value.

    Arguments:
        ndarray -- The ndarray to change.
        thickness -- Thickness of the border to change value.
        value -- Change into this value.

    Returns:
        An updated numpy 2D array object.
    """
    ndarray[0:thickness, :] = value
    ndarray[-thickness:, :] = value
    ndarray[:, 0:thickness] = value
    ndarray[:, -thickness:] = value
    return ndarray


def border_mask(shape, thickness):
    mask = np.zeros(shape, bool)
    set_value_to_border(mask, thickness, True)
    return mask


def generate_simple_boundaries(
    shape: tuple,
    outsides: int = 0,
    border: int = 1,
    boundary: Boundaries = None,
    **kwargs,
) -> Boundaries:
    """
    Generate the simplest boundary object.

    Arguments:
        shape -- Shape of the world, a 2D tuple.

    Keyword Arguments:
        outsides -- thickness of the outside area (default: {0}).
        border_thickness -- thickness of the border (default: {1}).
        boundary -- Boundaries object to update, or create a new one (default {None}).

    Returns:
        World's boundary object.
    """
    # 默认的边界是不同的
    if not boundary:
        boundary = Boundaries(shape=shape, **kwargs)
    if outsides > 0:
        outsides_ndarray = np.zeros(shape, dtype=bool)
        outsides_ndarray = set_value_to_border(
            outsides_ndarray, outsides, True
        )
        boundary.reset_boundary("outside", outsides_ndarray)
    if border > 0:
        border_ndarray = np.zeros(shape, dtype=bool)
        border_ndarray = set_value_to_border(
            border_ndarray, border + outsides, True
        )
        border_ndarray = set_value_to_border(border_ndarray, outsides, False)
        boundary.reset_boundary("boundaries", border_ndarray)
    boundary._update_interior()

    # Named special boundaries.
    left = np.zeros(shape, dtype=bool)
    left[
        outsides + border : -outsides - border, outsides : outsides + border
    ] = True
    right = np.zeros(shape, dtype=bool)
    right[
        outsides + border : -outsides - border,
        -outsides - border : shape[1] - outsides,
    ] = True
    # TODO Patches
    top = np.zeros(shape, dtype=bool)
    top[outsides : outsides + border, outsides : shape[1] - outsides] = True
    bottom = np.zeros(shape, dtype=bool)
    bottom[
        -outsides - border : shape[0] - outsides,
        outsides : shape[1] - outsides,
    ] = True

    boundary.change_special_boundary(name="left", patch=left)
    boundary.change_special_boundary(name="right", patch=right)
    boundary.change_special_boundary(name="top", patch=top)
    boundary.change_special_boundary(name="bottom", patch=bottom)
    boundary._update_all()
    return boundary


def simple_boundary_from(settings: dict):
    def parsing_shape(settings) -> tuple:
        y = settings.get("width", 8)
        x = settings.get("height", 6)
        shape = (x, y)
        return shape

    outside = settings.pop("outside", 0)
    border = settings.pop("border", 1)
    shape = parsing_shape(settings)
    boundary = generate_simple_boundaries(shape, outside, border)
    return boundary



#!/usr/bin/env python 3.11.0
# -*-coding:utf-8 -*-
# @Author  : Shuang (Twist) Song
# @Contact   : SongshGeo@gmail.com
# GitHub   : https://github.com/SongshGeo
# Website: https://cv.songshgeo.com/

from __future__ import annotations

import logging
import os
import threading
from collections import deque
from datetime import datetime
from functools import total_ordering
from numbers import Number
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    Self,
    Type,
    TypeAlias,
    Union,
)

import numpy as np
import pint
from agentpy.model import Model

from .patch import Patch
from .tools.func import make_list, wrap_opfunc_to

if TYPE_CHECKING:
    from .main import MainModel
    from .objects import BaseObject

MAXLENGTH = 5  # todo move it into system settings.

logger = logging.getLogger(__name__)

Data: TypeAlias = Union[Number, str, np.ndarray, Patch]
T: TypeAlias = datetime


# TODO feature: merge this
def setup_registry(registry):
    """set up the given registry for use with pint_xarray
    Namely, it enables ``force_ndarray_like`` to make sure results are always
    duck arrays.
    Parameters
    https://github.com/xarray-contrib/pint-xarray/blob/0cd77e80a07c58895d0851b6fdf1af646cb5bc19/pint_xarray/accessors.py
    ----------
    registry : pint.UnitRegistry
        The registry to modify
    """
    if not registry.force_ndarray and not registry.force_ndarray_like:
        registry.force_ndarray_like = True

    return registry


# https://github.com/xarray-contrib/pint-xarray/blob/0cd77e80a07c58895d0851b6fdf1af646cb5bc19/pint_xarray/accessors.py
# default_registry = setup_registry(pint.get_application_registry())


class VariablesRegistry:
    """Singleton Registry for each model."""

    _model: Dict[MainModel, VariablesRegistry] = {}
    _lock = threading.RLock()

    def __new__(cls: type[Self], model: MainModel) -> Self:
        instance = cls._model.get(model, None)
        if instance is None:
            instance = super().__new__(cls)
            with cls._lock:
                cls._model[model] = instance
        return instance

    def __init__(self, model: Model):
        self.model: Model = model
        self._variables: List[str] = []
        self._objs_registry: Dict[BaseObject, List[str]] = {}
        self._map: Dict[str, List[BaseObject]] = {}
        self._ln: Dict[str, str] = {}
        self._units: Dict[str, str] = {}
        self._data_types: Dict[str, Type] = {}
        self._checker: Dict[str, List[Callable]] = {}

    def __getitem__(self, owner) -> List[str]:
        return self._objs_registry.get(owner, [])

    def _check_type(
        self, var_name: str, dtype: Type, error_when: Optional[bool] = False
    ) -> bool:
        """Judge if the given variable's type matches its given data."""
        dtype_now = self._data_types.get(var_name, None)
        if dtype_now is None:
            flag = None
        elif dtype is dtype_now:
            flag = True
        else:
            flag = False
        if flag is error_when:
            raise ValueError(
                f"Failed in checking type of variable: '{dtype_now}', now, inputting {dtype}"
            )
        return flag

    def _is_registered(
        self, var_name: str, error_when: Optional[bool] = None
    ) -> bool:
        """Judge if the given variable is registered."""
        if var_name in self._variables:
            flag = True
        else:
            flag = False
        if flag is True and error_when is True:
            raise ValueError(f"Variable {var_name} already registered.")
        elif flag is False and error_when is False:
            raise ValueError(f"Variable {var_name} hasn't been registered")
        else:
            return flag

    def _has_variable(
        self, owner, var_name, error_when: Optional[bool] = None
    ) -> bool:
        """Judge if the given variable is belong to a given owner."""
        if owner in self._objs_registry:
            flag = var_name in self._objs_registry[owner]
        else:
            flag = False
            self._objs_registry[owner] = []
        if flag is True and error_when is True:
            raise ValueError(
                f"Variable '{var_name}' has already been registered in '{owner}'."
            )
        elif flag is False and error_when is False:
            raise ValueError(
                f"'{owner}' does not have a registered variable '{var_name}'."
            )
        else:
            return flag

    def _add_variable(
        self,
        var_name: str,
        long_name: Optional[str] = None,
        units: Optional[str] = None,
        dtype: Optional[Type] = None,
        check_func: Optional[Iterable[Callable]] = None,
    ) -> None:
        self._is_registered(var_name, error_when=True)
        self._check_type(var_name, dtype=dtype, error_when=False)
        self._variables.append(var_name)
        self._map[var_name] = list()
        self._ln[var_name] = long_name
        self._units[var_name] = units
        self._data_types[var_name] = dtype
        self._checker[var_name] = make_list(check_func)

    def register(
        self, owner: BaseObject, var_name: str, *args, **kwargs
    ) -> None:
        if not self._is_registered(var_name):
            self._add_variable(var_name, *args, **kwargs)
        self._has_variable(owner, var_name, error_when=True)
        self._objs_registry[owner].append(var_name)
        self._map[var_name].append(owner)

    def check_variable(self, var_name: str, value: Any) -> bool:
        """
        Check if the given value is valid data registered variable.

        Args:
            var_name (str): variable's name.
            value (Any): data to check.

        Returns:
            bool: if the value is valid data, returns True.
        """
        results = []
        results.append(self._is_registered(var_name))
        results.append(
            self._check_type(var_name, type(value), error_when=None)
        )
        for checker in self._checker[var_name]:
            results.append(checker(value))
        return all(results)

    def delete_variable(
        self, var_name: str, owner: Optional[BaseObject] = None
    ) -> None:
        """
        Delete a variable from its owner registry.

        Args:
            name (str): name of the variable to be deleted.
            owner (Optional, BaseObject): owner of this variable.
        """

        def remove_var_from_owner(owner):
            owned_vars = self._objs_registry[owner]
            owned_vars.remove(var_name)
            self._map[var_name].remove(owner)
            if len(owned_vars) == 0:
                del self._objs_registry[owner]

        self._is_registered(var_name, error_when=False)
        if owner is not None:
            self._has_variable(
                owner=owner, var_name=var_name, error_when=False
            )
            remove_var_from_owner(owner)
        else:
            for owner in tuple(self._map[var_name]):
                remove_var_from_owner(owner)
        # no owner has this variable, delete it from the registry.
        if len(self._map[var_name]) == 0 or owner is None:
            self._variables.remove(var_name)
            del self._ln[var_name]
            del self._units[var_name]
            del self._data_types[var_name]
            del self._checker[var_name]
            del self._map[var_name]


@total_ordering
class Variable:
    def __init__(
        self,
        now: Data,
        owner: Optional[BaseObject] = None,
        history: Iterable[Data] = None,
    ):
        # self.registry: VariablesRegistry = owner.registry
        self.data = now
        self.owner = owner
        self.history = history
        wrap_opfunc_to(self, "_data")

    # ----------------------------------------------------------------

    def __getattribute__(self, __name: str) -> Any:
        if __name.startswith("_"):
            return super().__getattribute__(__name)
        # using data explicit methods.
        elif callable(getattr(self._data, __name, None)):
            return getattr(self.data, __name)
        else:
            return super().__getattribute__(__name)

    def __lt__(self, other):
        return self.data < other

    def __eq__(self, other):
        return self.data == other

    def __repr__(self):
        return f"<[{self.owner}]Var: {self.data}>"

    # ----------------------------------------------------------------

    @property
    def data(self) -> Data:
        return self._data

    @data.setter
    def data(self, data: Data) -> None:
        # self.registry._check_dtype(data=data)
        self._data = data

    # ----------------------------------------------------------------

    # TODO processing before return
    def get_value(self) -> Data:
        return self.data


class FileVariable(Variable):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._filepath = None
        self._engine = None

    @property
    def path(self):
        return self._filepath

    def _file_is_existing(self, filepath: str) -> bool:
        return os.path.exists(filepath)



#!/usr/bin/env python 3.11.0
# -*-coding:utf-8 -*-
# @Author  : Shuang Song
# @Contact   : SongshGeo@gmail.com
# GitHub   : https://github.com/SongshGeo
# Website: https://cv.songshgeo.com/

"""
ABSESpy - Agent-based Social-ecological System framework in Python
Copyright (c) 2021-2023 Shuang Song

Documentation: https://agentpy.readthedocs.io/
Examples: https://agentpy.readthedocs.io/en/latest/model_library.html
Source: https://github.com/JoelForamitti/agentpy
"""

__all__ = [
    "__version__",
    "MainModel",
    "BaseHuman",
    "BaseNature",
    "HumanModule",
    "PatchModule",
    "Agent",
    "ActorsList",
    "AgentsContainer" "MyExperiment",
]
__version__ = "v0.1.0"

from .actor import Actor, link_to, perception
from .container import ActorsList, AgentsContainer
from .human import BaseHuman, HumanModule
from .main import MainModel
from .nature import BaseNature, PatchModule
from .objects import *



#!/usr/bin/env python 3.11.0
# -*-coding:utf-8 -*-
# @Author  : Shuang (Twist) Song
# @Contact   : SongshGeo@gmail.com
# GitHub   : https://github.com/SongshGeo
# Website: https://cv.songshgeo.com/

from abc import abstractmethod
from typing import Iterable, List, Optional, Set, Union

from agentpy.tools import AttrDict

from .bases import Mediator
from .log import Log
from .tools.func import iter_func, make_list

STATES = {
    -1: "Waiting",
    0: "new",
    1: "init",
    2: "ready",
    3: "complete",
}


class Component(Log):
    """
    The Base Component provides the basic functionality of storing a mediator's
    instance inside component objects.
    """

    def __init__(self, name: Optional[str] = None):
        Log.__init__(self, name=name)
        self._arguments: List[str] = []
        self._parameters: dict = AttrDict()

    @property
    def params(self) -> dict:
        return self._parameters

    @params.setter
    def params(self, my_params: dict) -> None:
        self.params.update(my_params)

    @property
    def arguments(self) -> List[str]:
        return self._arguments

    @arguments.setter
    def arguments(self, args: "str|Iterable[str]") -> None:
        arg_lst = make_list(args)
        self._arguments.extend(arg_lst)
        self._arguments = sorted(list(set(self.arguments)))

    @iter_func("modules")
    def _parsing_args(self) -> None:
        """
        Initialize arguments of the component then delete from param dictionary.

        Raises:
            KeyError: cannot find argument in parameters.
        """
        arguments = getattr(self.__class__, "args", [])
        self.arguments = arguments
        for arg in self.arguments:
            try:
                value = self.params[arg]
                self.__setattr__(arg, value)
                del self.params[arg]
            except KeyError:
                raise KeyError(f"arg '{arg}' not found in parameters.")

    @iter_func("modules")
    def parsing_params(self, params: dict) -> dict:
        """
        Parsing parameters belongs to this component.

        Args:
            params (dict): input settings.

        Returns:
            dict: unsolved parameters.
        """
        # select my params
        self.params = params.pop(self.name, {})
        # retrieve specific parameters and update
        for key in list(params.keys()):
            if key in self.params:
                value = params.pop(key)
                self.logger.debug(
                    f"Using {value} to update {self.params[key]} for '{key}'"
                )
                self.params[key] = value
        # setup arguments
        self._parsing_args()
        # solving parameters diversely
        self._after_parsing()
        # handle parameters defined by users
        self.handle_params()
        return params

    def report_parameters(
        self, params: Optional[dict] = None, as_string: Optional[bool] = False
    ) -> Union[dict, str]:
        # TODO: show parameter table.
        if len(params) == 0:
            return None
        if as_string:
            return str(list(params.keys()))
        else:
            return params.keys()

    @abstractmethod
    def _after_parsing(self):
        pass

    @abstractmethod
    def handle_params(self):
        """
        Handle parameters after loading.
        """
        pass

    @abstractmethod
    def initialize(self):
        """
        Initialization after handle parameters.
        """
        pass


class MainComponent(Component):
    _states = STATES

    def __init__(self, name: Optional[str] = None):
        Component.__init__(self, name=name)
        self._state = -1  # init state waiting
        self._mediator = Mediator()

    @property
    def mediator(self) -> Mediator:
        return self._mediator

    @mediator.setter
    def mediator(self, mediator) -> None:
        self._mediator = mediator

    @property
    def state(self) -> str:
        return self._states[self._state]

    @state.setter
    def state(self, code):
        if code not in self._states:
            raise ValueError(f"Invalid state {code}, valid: {self._states}!")
        elif code == self._state:
            raise ValueError(f"Setting state repeat: {self.state}!")
        elif code < self._state:
            raise ValueError(
                f"State cannot retreat from {self._states[code]} to {self.state}!"
            )
        else:
            self._state = code
        self.mediator.transfer_event(sender=self, event=self.state)

    def parsing_params(self, params: dict) -> dict:
        unsolved = super().parsing_params(params)
        finished = self.mediator.transfer_parsing(self, params)
        if finished:
            parameters = self.report_parameters(unsolved, as_string=True)
            message = f"Unsolved parameters: {parameters}."
            self.mediator.logging(
                message, condition=len(unsolved) > 0, level="warning"
            )
        return unsolved

    @iter_func("modules")
    def _after_parsing(self):
        pass



#!/usr/bin/env python 3.11.0
# -*-coding:utf-8 -*-
# @Author  : Shuang (Twist) Song
# @Contact   : SongshGeo@gmail.com
# GitHub   : https://github.com/SongshGeo
# Website: https://cv.songshgeo.com/

import logging
from collections.abc import Iterable
from numbers import Number
from typing import (
    Any,
    Callable,
    Dict,
    Iterator,
    List,
    Optional,
    Self,
    TypeAlias,
    Union,
    overload,
)

import numpy as np
from agentpy import AgentList

from .actor import Actor
from .tools.func import make_list, norm_choice

logger = logging.getLogger("__name__")

Selection: TypeAlias = Union[str, Iterable[bool]]


class ActorsList(AgentList):
    def __repr__(self):
        results = []
        for k, v in self.to_dict().items():
            results.append(f"({len(v)}){k}")
        return f"<ActorsList: {'; '.join(results)}>"

    def __getattr__(self, name: str) -> np.ndarray:
        """Return callable list of attributes"""
        if name[0] == "_":  # Private variables are looked up normally
            super().__getattr__(name)
        elif name in self.__dir__():
            return super().__getattr__(name)
        elif hasattr(getattr(self[0], name, None), "__call__"):
            return super().__getattr__(name)
        else:
            return ActorsList.array(self, name)

    def __iter__(self) -> Iterator[Actor]:
        return super().__iter__()

    def __eq__(self, other: Iterable) -> bool:
        if not self._is_same_length(other):
            return False
        else:
            return all([actor in other for actor in self])

    @overload
    def __getitem__(self, other: int) -> Actor:
        ...

    @overload
    def __getitem__(self, index: slice) -> Self:
        ...

    def __getitem__(self, index):
        results = super().__getitem__(index)
        if isinstance(index, slice):
            return ActorsList(self.model, results)
        else:
            return results

    def _is_same_length(
        self, length: Iterable[Any], rep_error: bool = False
    ) -> bool:
        """Check if the length of input is as same as the number of actors."""
        if not hasattr(self, "__len__"):
            raise ValueError(f"{type(length)} object is not iterable.")
        elif not length.__len__() == self.__len__():
            if rep_error:
                raise ValueError(
                    f"Length of the input {len(length)} mismatch {len(self)} actors."
                )
            return False
        else:
            return True

    def to_dict(self) -> Dict[str, Self]:
        """
        Convert all actors in this list to a dictionary.

        Returns:
            Dict[str, Self]: key is the breed of actors, and values are corresponding actors.
        """
        dic = {}
        for actor in self.__iter__():
            breed = actor.breed
            if breed not in dic:
                dic[breed] = ActorsList(self.model, [actor])
            else:
                dic[breed].append(actor)
        return dic

    def select(self, selection: Selection) -> Self:
        """Returns a new :class:`ActorList` based on `selection`.

        Arguments:
            selection (list of bool): List with same length as the agent list.
                Positions that return True will be selected.
        """
        if isinstance(selection, (str, dict)):
            bool_ = [actor.selecting(selection) for actor in self]
        elif isinstance(selection, (list, tuple, np.ndarray)):
            bool_ = make_list(selection)
        else:
            raise TypeError(f"Invalid selection type {type(selection)}")
        selected = [a for a, s in zip(self, bool_) if s]
        return ActorsList(self.model, selected)

    def now(self) -> Self:
        """Only select actors who is already setup on the earth."""
        return self.select(self.on_earth)

    def ids(self, ids: Iterable[int]) -> List[Actor]:
        """
        Select by a `ids` list of Agent.

        Args:
            ids (iterable): an iterable id list. List[id], ID is an attr of agent obj.

        Returns:
            ActorList: A subset of origin agents list.
        """
        ids = make_list(ids)
        agents = self.select([agent.id in ids for agent in self])
        return agents

    def random_choose(
        self,
        p: Optional[Iterable[float]] = None,
        size: int = 1,
        replace: bool = True,
    ) -> Union[Actor, Self]:
        if size == 1:
            return norm_choice(self, p=p, replace=replace)
        elif size > 1:
            chosen = norm_choice(self, p=p, size=size, replace=replace)
            return ActorsList(self.model, chosen)

    def better(
        self, metric: str, than: Optional[Union[Number, Actor]] = None
    ) -> Self:
        metrics = self.array(attr=metric)
        if than is None:
            return self.select(metrics == max(metrics))
        elif isinstance(than, Number):
            return self.select(metrics > than)
        elif isinstance(than, Actor):
            diff = self.array(metric) - getattr(than, metric)
            return self.select(diff > 0)

    def update(self, attr: str, values: Iterable[any]) -> None:
        self._is_same_length(values, rep_error=True)
        [agent.__setattr__(attr, val) for agent, val in zip(self, values)]

    def split(self, where: Iterable[int]) -> np.ndarray:
        """
        split agents into N+1 groups.

        Args:
            where (Iterable[int]): indexes [size=N] denotes where to split.

        Returns:
            np.ndarray: N+1 groups: agents array
        """
        to_split = np.array(self)
        return np.hsplit(to_split, where)

    def array(
        self, attr: str, how: "int|str" = "attr", *args, **kwargs
    ) -> np.ndarray:
        if how == "attr":
            results = [getattr(actor, attr) for actor in self]
        # elif how == "loc":
        #     results = self.loc(attr)
        # elif how == "mine":
        #     results = self.mine(attr, *args, **kwargs)
        return np.array(results)

    def position_to_coord(self, x_coords, y_coords):
        # TODO move this to actor's property
        X, Y = np.meshgrid(x_coords, y_coords)
        lon = [X[x, y] for x, y in self.pos]
        lat = [Y[x, y] for x, y in self.pos]
        return lon, lat

    def trigger(self, func_name: str, *args, **kwargs) -> np.ndarray:
        results = []
        for actor in self.__iter__():
            func = getattr(actor, func_name)
            results.append(func(*args, **kwargs))
        return np.array(results)



#!/usr/bin/env python 3.11.0
# -*-coding:utf-8 -*-
# @Author  : Shuang Song
# @Contact   : SongshGeo@gmail.com
# GitHub   : https://github.com/SongshGeo
# Website: https://cv.songshgeo.com/

from __future__ import annotations

import logging
from collections import deque
from typing import (
    Any,
    Callable,
    Deque,
    Dict,
    Iterable,
    List,
    Optional,
    Set,
    Union,
)

import pandas as pd
from agentpy import AttrDict
from agentpy.model import Model
from agentpy.objects import Object
from prettytable import PrettyTable

from abses.time import TimeDriver
from abses.tools.func import make_list

from .bases import Observer
from .log import Log
from .variable import MAXLENGTH, Data, Variable, VariablesRegistry

logger = logging.getLogger(__name__)


class BaseObj(Observer, Log, Object):
    def __init__(
        self,
        model: Model,
        observer: Optional[bool] = True,
        name: Optional[str] = None,
    ):
        Object.__init__(self, model=model)
        Log.__init__(self, name=name)
        self._registry: VariablesRegistry = VariablesRegistry(model=model)
        self._vars_history: Dict[str, Deque[Variable]] = AttrDict()
        self._time: TimeDriver = TimeDriver(model=model)
        self.glob_vars: List[str] = []
        self._recording: Set[str] = set()
        self._reporting: Set[str] = set()
        if observer:
            model.attach(self)

    def __getattr__(self, __name: str) -> Any:
        # if __name in self.vars:
        #     now = super().__getattribute__(__name)
        #     var = self.to_variable(__name, now)
        #     return var.get_value()
        return super().__getattribute__(__name)

    @property
    def vars(self) -> List[str]:
        return self._registry[self]

    @property
    def time(self) -> TimeDriver:
        return self._time

    @property
    def output(self) -> pd.DataFrame:
        df = pd.DataFrame(self.log, index=self.time.time[:-1])
        return df

    @property
    def recording(self) -> Set[str]:
        return self._recording

    @recording.setter
    def recording(self, variables: Union[str, Iterable[str]]) -> None:
        self._recording = self.recording.union(set(make_list(variables)))

    @property
    def reporting(self) -> Set[str]:
        return self._reporting

    @reporting.setter
    def reporting(self, variables: Union[str, Iterable[str]]) -> None:
        self._reporting = self.reporting.union(make_list(variables))

    def _when_time_go(self):
        self.record(self._recording)
        for var in self._registry[self]:
            value_now = self.__getattr__(var)
            self._vars_history[var].append(value_now)

    def to_variable(self, name, now: Optional[Data] = None) -> Data:
        self._registry.check_variable(name, value=now)
        history = self._vars_history[name]
        return Variable(owner=self, history=history, now=now)

    def register_a_var(
        self,
        name: str,
        init_data: Optional[Data] = None,
        long_name: Optional[str] = None,
        units: Optional[str] = None,
        check_func: Optional[Iterable[Callable]] = None,
    ) -> Variable:
        dtype = type(init_data)
        self._registry.register(
            owner=self,
            var_name=name,
            long_name=long_name,
            units=units,
            dtype=dtype,
            check_func=check_func,
        )
        self._vars_history[name] = deque([], maxlen=MAXLENGTH)
        self.__setattr__(name, init_data)
        var = self.to_variable(name=name, now=init_data)
        return var

    def report(
        self,
        max_width: int = 30,
        decimal: int = 4,
    ) -> PrettyTable:
        table = PrettyTable()
        table.field_names = ["Variable", "Value"]
        for attr, val in self.__dict__.items():
            if attr in ("id", "type"):
                continue
            if attr.startswith("_"):
                continue
            table.add_row([attr, val])
        table.max_width = max_width
        table.title = f"{self.type} {self.id}:"
        table.float_format = f".{decimal}"
        return table



#!/usr/bin/env python 3.11.0
# -*-coding:utf-8 -*-
# @Author  : Shuang (Twist) Song
# @Contact   : SongshGeo@gmail.com
# GitHub   : https://github.com/SongshGeo
# Website: https://cv.songshgeo.com/

import numpy as np


# 随机选择一个；根据偏好学习；就是得分 -> 概率
def fermi_ruler(profits_diff, k=0.5):
    # G. Szabó, G. Fáth, "Evolutionary games on graphs," Phys. Rep. 446, 97 2007
    # 1 / (1 + np.exp((n_{sx} - k_h) / k))
    # n_{sx}: 周围邻居中采取与个体x相同策略的邻居数目
    # k_h: degree/2
    # 可以在群体里选择一部分人（从众个体）
    # 可以在群体里定义一部分收益个体
    return 1 / (1 + np.exp(profits_diff / k))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['abses', 'abses.algorithms', 'abses.tools']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'abses',
    'version': '0.1.1',
    'description': 'ABSESpy makes it easier to build artificial Social-ecological systems with real GeoSpatial datasets and fully incorporate human behaviour.',
    'long_description': '![pixel_abses2](https://songshgeo-picgo-1302043007.cos.ap-beijing.myqcloud.com/uPic/pixel_abses2.svg)\n\n[![license](https://img.shields.io/github/license/songshgeo/absespy)](http://www.apache.org/licenses/) ![downloads](https://img.shields.io/github/downloads/songshgeo/absespy/total) ![codesize](https://img.shields.io/github/languages/code-size/songshgeo/absespy) ![tag](https://img.shields.io/github/v/tag/songshgeo/absespy)\n[![github](https://img.shields.io/badge/Website-SongshGeo-brightgreen.svg)](https://cv.songshgeo.com/) ![stars](https://img.shields.io/github/stars/songshgeo/absespy?style=social) [![twitter](https://img.shields.io/twitter/follow/shuangsong11?style=social)](https://twitter.com/shuangsong11)\n\n<!-- Language: [English Readme](#) | [简体中文](README_ch) -->\n\n`ABSESpy` makes it easier to build artificial **[Social-ecological systems](https://songshgeo.github.io/ABSESpy/docs/about/)** with real **GeoSpatial datasets** and fully incorporate **human behaviour**.\n\n## Why `ABSESpy`?\n\n**Several characteristics of Agent-Based model (ABM) make it an essential method for social-ecological systems (SES) research:**\n\n1. It focuses on the change of an SES over time from mutual adaptations of agents and their environments.\n2. its ability to generate emergent system-level outcomes from micro-level interactions and macro-level feedback.\n3. its ability to represent the diversity and heterogeneity of human and non-human actors and the spatial characteristics of an SES ...\n\nHowever, there is currently no modelling framework that **combines geo-spatial data and actor behaviour** (`actor` is the term for agents in the SES framework) well. `ABSESpy` is designed for spatial modelling that **couples human and nature** by:\n\n- Modelling environment for agents with **[geo data](https://songshgeo.github.io/ABSESpy/tutorial/notebooks/nature/geodata/)**: `Shapefile`, `GeoTiff`, `NetCDF`.\n- Modelling **[human behaviour](https://songshgeo.github.io/ABSESpy/tutorial/notebooks/human/CCR_example/)** of agents with [cognition, contagion and responses](https://songshgeo.github.io/ABSESpy/docs/background/#human-behaviour-framework).\n- Easily manage all [parameters, arguments](https://songshgeo.github.io/ABSESpy/tutorial/notebooks/parameters/), and variables with a `yaml` settings file.\n\n## Install\n\nInstall with pip or your favourite PyPI package manager.\n```\npip install abses\n```\n\n## Basic usage & Documents\n\nYou can see how to use `ABSESpy` in these simple [tutorials](https://songshgeo.github.io/ABSESpy/tutorial/user_guide/):\n\n1. [Organize an Agent-based model](https://songshgeo.github.io/ABSESpy/tutorial/notebooks/model/) and [easily manage parameters](https://songshgeo.github.io/ABSESpy/tutorial/notebooks/parameters/).\n2. Using [geo-spatial data](https://songshgeo.github.io/ABSESpy/tutorial/notebooks/nature/geodata/) as the environment.\n3. Simply applying a [human behavior framework](https://songshgeo.github.io/ABSESpy/tutorial/notebooks/human/CCR_example/).\n\n<img src="https://songshgeo-picgo-1302043007.cos.ap-beijing.myqcloud.com/uPic/2jFmIS.jpg" alt="Drawing" style="width: 400px;"/>\n\nAccess the [full Documentation here](https://songshgeo.github.io/ABSESpy/).\n\n## Get in touch\n\n- **For enthusiastic developers and contributors**, all contributions, bug reports, bug fixes, documentation improvements, enhancements, and ideas are welcome.\n- **For SES researchers**, welcome to use this package in social-ecological system (SES) studies. If you have a model published, feel free to contribute it to our model library through [mailing list](https://groups.google.com/g/absespy).\n\nIf you need any help when using `ABSESpy`, don\'t hesitate to get in touch with us through:\n\n- Ask usage questions ("How to do?") on\xa0[_GitHub_\xa0Discussions](https://github.com/SongshGeo/ABSESpy/discussions).\n- Report bugs, suggest features or view the source code\xa0[on\xa0_GitHub_](https://github.com/SongshGeo/ABSESpy/issues).\n- For less well-defined questions or ideas or to announce other projects of interest to xarray users, use the\xa0[mailing list](https://groups.google.com/g/absespy).\n\n## License\n\nCopyright 2023, `ABSESpy` [Shuang Song](https://cv.songshgeo.com/)\n\nLicensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the License. You may obtain a copy of the License at\n\n[https://www.apache.org/licenses/LICENSE-2.0](https://www.apache.org/licenses/LICENSE-2.0)\n\nUnless required by applicable law or agreed to in writing, software distributed under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the specific language governing permissions and limitations under the License.\n\n`ABSESpy` bundles portions of `AgentPy`, `pandas`, `NumPy` and `Xarray`; the full text of these licenses is included in the licenses directory.\n',
    'author': 'Shuang Song',
    'author_email': 'songshgeo@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.11,<4.0',
}


setup(**setup_kwargs)



#!/usr/bin/env python 3.11.0
# -*-coding:utf-8 -*-
# @Author  : Shuang (Twist) Song
# @Contact   : SongshGeo@gmail.com
# GitHub   : https://github.com/SongshGeo
# Website: https://cv.songshgeo.com/

from __future__ import annotations

import logging
import threading
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Iterable,
    Optional,
    Self,
    Type,
    Union,
)

from agentpy import AttrDict

from .actor import Actor
from .sequences import ActorsList, Selection
from .tools.func import make_list, norm_choice

if TYPE_CHECKING:
    from .main import MainModel

logger = logging.getLogger("__name__")


class AgentsContainer(AttrDict):
    """Singleton AgentsContainer for each model."""

    _models: Dict[MainModel, AgentsContainer] = {}
    _lock = threading.RLock()

    def __new__(cls: type[Self], model: MainModel) -> Self:
        instance = cls._models.get(model, None)
        if instance is None:
            instance = super().__new__(cls)
            with cls._lock:
                cls._models[model] = instance
        return instance

    def __init__(self, model: MainModel):
        self._model = model
        self._breeds = {}

    def __len__(self):
        return len(self.to_list())

    def __repr__(self):
        # "<ActorsList: >"
        rep = self.to_list().__repr__()[13:-1]
        return f"<AgentsContainer: {rep}>"

    def __getattr__(self, name: str) -> Any:
        if name[0] == "_":
            return super().__getattr__(name)
        elif name in self._breeds:
            return self.to_list(name)
        else:
            return super().__getattr__(name)

    def __contains__(self, name):
        return name in self.to_list()

    @property
    def breeds(self):
        return tuple(self._breeds.keys())

    def _register(self, actor: Actor) -> None:
        if actor.breed not in self._breeds:
            self._breeds[actor.breed] = type(actor)
            self[actor.breed] = set()

    def create(
        self, breed_cls: Type[Actor], n: int = 1
    ) -> Union[Actor, ActorsList]:
        agents = ActorsList(self._model, objs=n, cls=breed_cls)
        self.add(agents, register=True)
        if n == 1:
            return agents[0]
        else:
            return agents

    # def create_from(self, breeds: dict):
    #     for breed_cls, n in breeds.items():
    #         self.create_agents(breed_cls, n)

    def to_list(
        self, breeds: Optional[Union[str, Iterable[str]]] = None
    ) -> ActorsList:
        if breeds is None:
            breeds = self.breeds
        agents = ActorsList(self._model)
        for breed in make_list(breeds):
            agents.extend(self[breed])
        return agents

    def add(
        self,
        agents: Union[Actor, ActorsList, Iterable[Actor]] = None,
        register: bool = False,
    ) -> None:
        dic = ActorsList(self._model, make_list(agents)).to_dict()
        for k, actors_lst in dic.items():
            if k not in self.breeds:
                if register:
                    self._register(actors_lst[0])
                else:
                    raise TypeError(f"'{k}' not registered.")
            self[k] = self[k].union(actors_lst)

    def remove(self, agent: Actor) -> None:
        self[agent.breed].remove(agent)
        if agent.on_earth is True:
            agent.here.remove(agent)

    def select(self, selection: Selection) -> ActorsList:
        return self.to_list().select(selection)


def apply_agents(func) -> callable:
    """
    Apply this method to all agents of model if input agents argument is None.

    Args:
        manipulate_agents_func (wrapped function): should be a method of module.
    """

    def apply(self, agents=None, *args, **kwargs):
        if agents is None:
            agents = self.model.agents
            results = agents.apply(func, sender=self, *args, **kwargs)
        else:
            results = func(self, agents=agents, *args, **kwargs)
        return results

    return apply



#!/usr/bin/env python 3.11.0
# -*-coding:utf-8 -*-
# @Author  : Shuang Song
# @Contact   : SongshGeo@gmail.com
# GitHub   : https://github.com/SongshGeo
# Website: https://cv.songshgeo.com/

import logging
import os

import dill as pickle

logger = logging.getLogger(__name__)


class Folder:
    def __init__(self, base):
        base = self.base_is_valid(base)
        self._base = base

    @property
    def path(self):
        return os.path.join(os.getcwd(), self.base)

    @property
    def base(self):
        return self._base

    def base_is_valid(self, base, create: bool = False):
        if base is None:
            base = ""
            logger.warning(
                f"Setup with NO location specified, will save under: {os.getcwd()}"
            )
        if create:
            self.create_path()
        return base

    def create_path(self, folder: str = "") -> str:
        path = os.path.join(self.path, folder)
        if not os.path.exists(path):
            os.mkdir(path)
        return path

    def to_pickle(self):
        file_path = os.path.join(self.path, f"{self.name}.pkl")
        with open(file_path, "wb") as pkl:
            pickle.dump(obj=self, file=pkl)
        logger.info(f"Model {self.name} saved under {self.path}.")

    pass



#!/usr/bin/env python 3.11.0
# -*-coding:utf-8 -*-
# @Author  : Shuang Song
# @Contact   : SongshGeo@gmail.com
# GitHub   : https://github.com/SongshGeo
# Website: https://cv.songshgeo.com/

import logging
import logging.config
import os

import yaml

from .tools.func import iter_func

CONFIG_PATH = os.path.join(os.getcwd(), "config/log.yaml")
with open(CONFIG_PATH, "rt") as f:
    config_data = yaml.safe_load(f.read())
    logging.config.dictConfig(config_data)


class StreamToLogger(object):
    """
    Fake file-like stream object that redirects writes to a logger instance.
    """

    def __init__(self, logger, log_level=logging.INFO):
        self.logger = logger
        self.log_level = log_level
        self.linebuf = ""

    def write(self, buf):
        for line in buf.rstrip().splitlines():
            self.logger.log(self.log_level, line.rstrip())


def clean_logs():
    handlers = config_data.get("handlers")
    removed_handlers = []
    for handler, configs in handlers.items():
        filename = configs.get("filename")
        if not filename:
            continue
        file_path = os.path.join(os.getcwd(), filename)
        open(file_path, "w").close()
        removed_handlers.append(handler)
    logging.warning(f"Logs of {removed_handlers} cleared.")


# TODO change a specific log for experiments
class Log(object):
    def __init__(self, name):
        if name is None:
            name = self.__class__.__name__.lower()
        self._name = name
        self._log = logging.getLogger(name)
        self._log_flag = True

    @property
    def name(self) -> str:
        return self._name

    @property
    def logger(self):
        return self._log

    @property
    def log_flag(self):
        return self._log_flag

    @iter_func("modules")
    def close_log(self):
        self._log_flag = False



#!/usr/bin/env python 3.11.0
# -*-coding:utf-8 -*-
# @Author  : Shuang Song
# @Contact   : SongshGeo@gmail.com
# GitHub   : https://github.com/SongshGeo
# Website: https://cv.songshgeo.com/

import logging
import os
from typing import Iterable

import xarray as xr
import yaml

from .func import make_list

logger = logging.getLogger(__name__)


def is_valid_yaml(path: any) -> bool:
    """
    Judge if the input is a valid yaml file path.

    Args:
        path (any): element to judge.

    Returns:
        bool: if is a valid yaml file path, returns True.
    """
    if not isinstance(path, str):
        return False
    elif not path.endswith(".yaml"):
        return False
    elif not os.path.isfile(path):
        raise ValueError(
            f"{path} contains '.yaml' but is not a valid yaml file path."
        )
    else:
        return True


def redirect_yaml(param: dict) -> None:
    """
    Clean dictionary, if valid-yaml file path exists, replace it with a new dictionary.

    Args:
        param (dict): dictionary to clean.
    """
    for key, val in param.items():
        if isinstance(val, dict):
            redirect_yaml(val)
        elif is_valid_yaml(val):
            param[key] = read_yaml(val, True)


def read_yaml(path: str, nesting: bool = True) -> dict:
    """
    Load parameters from yaml settings file.

    Arguments:
        yaml_file -- File path.
        detecting_nesting (bool, optional): wether to detect nesting `yaml` files. Defaults to True.

    Raises:
        KeyError: Cannot find a param by key.
    """
    # Read yaml file
    with open(path, "r", encoding="utf-8") as file:
        yaml_data = file.read()
        params = yaml.load(yaml_data, Loader=yaml.FullLoader)
        file.close()
    if nesting:
        redirect_yaml(params)
    return params


def get_files_in_folder(
    dir_path,
    suffix: "str|Iterable[str]" = "",
    iteration: bool = True,
    full_path: bool = False,
) -> list[str]:
    """
    Read all files in a specific path.

    Args:
        dir_path (str): folder path.
        postfix (str|Iterable[str], optional): valid postfix. Defaults to "".
        iteration (bool, optional): if sub-folders include. Defaults to True.
        full_path (bool, optional): return full path of the files, or just the filename. Defaults to False.

    Raises:
        FileNotFoundError: Invalid dir_path.

    Returns:
        list[str]: list of filenames (default),
        or full path of the files (if full_path) is True.
    """
    if not os.path.isdir(dir_path):
        raise FileNotFoundError(f"{dir_path} not a valid folder.")
    paths = []
    valid_postfix = make_list(suffix)
    if iteration:
        for root, _, files in os.walk(dir_path, topdown=False):
            for filename in files:
                validation = [
                    filename.endswith(suffix) for suffix in valid_postfix
                ]
                if not any(validation):
                    continue
                if full_path:
                    path = os.path.join(root, filename)
                    paths.append(path)
                else:
                    paths.append(filename)
    else:
        filename = os.listdir(dir_path)
        paths.append(os.path.join(dir_path, filename))
    return paths



#!/usr/bin/env python 3.11.0
# -*-coding:utf-8 -*-
# @Author  : Shuang (Twist) Song
# @Contact   : SongshGeo@gmail.com
# GitHub   : https://github.com/SongshGeo
# Website: https://cv.songshgeo.com/

from __future__ import annotations

import threading
from collections import deque
from typing import TYPE_CHECKING, Any, Deque, Dict, List, Optional

from pandas import Period

from .tools.func import wrap_opfunc_to
from .variable import MAXLENGTH

if TYPE_CHECKING:
    from .main import MainModel

DEFAULT_START = "2000-01-01 00:00:00"
DEFAULT_END = "2023-01-01 00:00:00"
DEFAULT_FREQ = "Y"


class TimeDriverManager:
    def __init__(self, model: MainModel) -> None:
        settings = model.params.get("time", {})
        start = self._parsing_settings(
            settings, key="start", defaults=DEFAULT_START
        )
        self._end = self._parsing_settings(
            settings, key="end", defaults=DEFAULT_END
        )
        self._data: Period = start
        self._time: List[Period] = [start]
        self._history: Deque[Period] = deque([], maxlen=MAXLENGTH)
        self._model: MainModel = model
        self._settings: Dict[str, Any] = model.params.get("time", {})
        # instances operational func using the property '_data'
        wrap_opfunc_to(self, "_data")

    def __getattribute__(self, __name: str) -> Any:
        # private attributes of TimeDriverManager
        if __name.startswith("_"):
            return super().__getattribute__(__name)
        # using data explicit methods.
        elif hasattr(self._data, __name):
            return getattr(self._data, __name)
        # accessible methods and attributes of TimeDriver
        else:
            TimeDriver._select_manager(self._model)
            return getattr(TimeDriver, __name)

    # cls(2000, 'Y') < Period(2001, 'Y')
    def __lt__(self, other):
        return self._data < other

    # cls(2000, 'Y') == Period(2000, 'Y')
    # cls(2000, 'Y') != Period(2000, 'M')
    def __eq__(self, other):
        return self._data == other

    def __str__(self) -> str:
        """[A-DEC]2000-2001"""
        start = self._time[0]
        end = self._time[-1]
        return f"[{self.freqstr}]{start}-{end}"

    def __repr__(self) -> str:
        return self._data.__str__()

    @staticmethod
    def _parsing_settings(
        settings: Dict[str, Any], key: str, defaults: Optional[str]
    ) -> Period:
        freq = settings.get("freq", DEFAULT_FREQ)
        values = settings.get(key, defaults)
        if not isinstance(values, dict):
            return Period(value=values, freq=freq)
        else:
            return Period(freq=freq, **values)


class TimeDriver(Period):
    _manager: TimeDriverManager = None
    _model: Dict[int, TimeDriverManager] = {}
    _lock = threading.RLock()

    # 单例模式：https://www.jb51.net/article/202178.htm
    def __new__(cls, model: MainModel) -> TimeDriverManager:
        """A Singleton wrapped class, each model has its own driver.
        This class has NO instance, but init a TimeDriverManager.
        Each model can only store one initialized TimeDriverManager instance.
        """
        # if this is the first time to initialize.
        if cls._model.get(model) is None:
            driver = TimeDriverManager(model)
            with cls._lock:
                cls._model[model] = driver
        # if this model has a TimeDriverManager.
        else:
            driver = cls._model[model]
        return driver

    @classmethod
    def _select_manager(cls, model: int) -> None:
        """Switch to current model's manager."""
        with cls._lock:
            cls.manager = cls._model[model]

    @classmethod
    @property
    def period(cls) -> Period:
        return cls.manager._data

    @classmethod
    @property
    def history(cls) -> Deque[Period]:
        return cls.manager._history

    @classmethod
    @property
    def time(cls) -> List[Period]:
        return cls.manager._time

    @classmethod
    @property
    def manager(cls) -> TimeDriverManager:
        return cls._manager

    @classmethod
    def update(cls, steps: Optional[int] = 1) -> Period:
        """
        Update the time period.

        Args:
            steps (Optional[int], optional): how many time steps to update. Defaults to 1.

        Returns:
            Period: after updating, the current time period.
        """
        for _ in range(steps):
            cls.history.append(cls.period)
            cls.manager._data = cls.period + 1
            cls.time.append(cls.period)
        return cls.manager._data
